# -*- coding: utf-8 -*-
#
# pointpats documentation build configuration file, created by
# sphinx-quickstart on Nov 12 16:37:22 2018.
#
# This file is execfile() with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import sys, os
import sphinx_bootstrap_theme


sys.path.insert(0, os.path.abspath('../'))

import pointpats


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [#'sphinx_gallery.gen_gallery',
              'sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.viewcode',
              'sphinxcontrib.bibtex',
              'sphinx.ext.mathjax',
              'sphinx.ext.doctest',
              'sphinx.ext.intersphinx',
              'numpydoc',
              'matplotlib.sphinxext.plot_directive']



# sphinx_gallery_conf = {
#      # path to your examples scripts
#      'examples_dirs': '../examples',
#      # path where to save gallery generated examples
#      'gallery_dirs': 'auto_examples',
#      'backreferences_dir': False,
# }


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'pointpats'
copyright = '2018-, pysal developers'
author = 'pysal developers'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version.
version = pointpats.__version__
release = pointpats.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'tests/*']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

bibtex_bibfiles = ['_static/references.bib']

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'bootstrap'
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()
html_title = "%s v%s Manual" % (project, version)

# (Optional) Logo. Should be small enough to fit the navbar (ideally 24x24).
# Path should be relative to the ``_static`` files directory.
#html_logo = "_static/images/CGS_logo.jpg"
#html_logo = "_static/images/CGS_logo_green.png"
#html_logo = "_static/images/pysal_logo_small.jpg"
html_favicon = "_static/images/pysal_favicon.ico"


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {

    # Navigation bar title. (Default: ``project`` value)
    'navbar_title': "pointpats",

    # Render the next and previous page links in navbar. (Default: true)
    'navbar_sidebarrel': False,

    # Render the current pages TOC in the navbar. (Default: true)
    #'navbar_pagenav': True,
    #'navbar_pagenav': False,

    # No sidebar
    'nosidebar': True,

    # Tab name for the current pages TOC. (Default: "Page")
    #'navbar_pagenav_name': "Page",

    # Global TOC depth for "site" navbar tab. (Default: 1)
    # Switching to -1 shows all levels.
    'globaltoc_depth': 2,

    # Include hidden TOCs in Site navbar?
    #
    # Note: If this is "false", you cannot have mixed ``:hidden:`` and
    # non-hidden ``toctree`` directives in the same page, or else the build
    # will break.
    #
    # Values: "true" (default) or "false"
    'globaltoc_includehidden': "true",

    # HTML navbar class (Default: "navbar") to attach to <div> element.
    # For black navbar, do "navbar navbar-inverse"
    #'navbar_class': "navbar navbar-inverse",

    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    'navbar_fixed_top': "true",


    # Location of link to source.
    # Options are "nav" (default), "footer" or anything else to exclude.
    'source_link_position': 'footer',

    # Bootswatch (http://bootswatch.com/) theme.
    #
    # Options are nothing (default) or the name of a valid theme
    # such as "amelia" or "cosmo", "yeti", "flatly".
    'bootswatch_theme': "yeti",

    # Choose Bootstrap version.
    # Values: "3" (default) or "2" (in quotes)
    'bootstrap_version': "3",

    'navbar_links': [
                     ("Installation", "installation"),
                     ("API", "api"),
                     ("References", "references"),
                     ],

}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}
# html_sidebars = {'sidebar': ['localtoc.html', 'sourcelink.html', 'searchbox.html']}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'pointpatsdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'pointpats.tex', u'pointpats Documentation',
     u'pysal developers', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'pointpats', u'pointpats Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'pointpats', u'pointpats Documentation',
     author, 'pointpats', 'One line description of project.',
     'Miscellaneous'),
]

# -----------------------------------------------------------------------------
# Napoleon configuration
# -----------------------------------------------------------------------------
# numpydoc_show_class_members = True
# numpydoc_class_members_toctree = False
#
# napoleon_use_ivar = True

# -----------------------------------------------------------------------------
# Autosummary
# -----------------------------------------------------------------------------

# Generate the API documentation when building
autosummary_generate = True

# avoid showing members twice
numpydoc_show_class_members = False
numpydoc_use_plots = True

# automatically document class members
autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'inherited-members': True
}
# display the source code for Plot directive
plot_include_source = True

def setup(app):
    app.add_css_file("pysal-styles.css")

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ('https://docs.python.org/3', None),
                       'numpy': ('https://docs.scipy.org/doc/numpy', None),
                       'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
                       'libpysal': ('https://pysal.org/libpysal/', None),
                       'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
                       'matplotlib':("https://matplotlib.org/", None),
                       'opencv-contrib-python':("https://docs.opencv.org/3.4/index.html", None),
                       'KDEpy':("https://kdepy.readthedocs.io/en/latest/", None),
                       'statsmodels':("https://www.statsmodels.org/stable/", None),
                       }





import numpy
from scipy import spatial
from pointpats import distance_statistics as ripley, geometry, random
from libpysal.cg import alpha_shape_auto
import shapely
import warnings
import pytest

points = numpy.asarray(
    [
        [66.22, 32.54],
        [22.52, 22.39],
        [31.01, 81.21],
        [9.47, 31.02],
        [30.78, 60.10],
        [75.21, 58.93],
        [79.26, 7.68],
        [8.23, 39.93],
        [98.73, 77.17],
        [89.78, 42.53],
        [65.19, 92.08],
        [54.46, 8.48],
    ]
)

tree = spatial.KDTree(points)

chull = spatial.ConvexHull(points)
ashape = alpha_shape_auto(points)

bbox = numpy.asarray((*points.min(axis=0), *points.max(axis=0)))

support = numpy.linspace(0, 100, num=15)

d_self = spatial.distance.pdist(points)
D_self = spatial.distance.squareform(d_self)
try:
    numpy.random.seed(2478879)
    random_pattern = random.poisson(bbox, size=500)
    D_other = spatial.distance.cdist(points, random_pattern)
except:
    # will cause failures in all ripley tests later from NameErrors about D_other
    # If D_other is missing, then test_simulate should also fail.
    pass


def test_primitives():
    area_bbox = (bbox[2] - bbox[0]) * (bbox[3] - bbox[1])
    assert area_bbox == geometry.area(bbox)
    area_chull = chull.volume
    assert area_chull == geometry.area(chull)
    area_pgon = geometry.area(ashape)
    assert area_pgon == ashape.area
    assert area_pgon == geometry.area(ashape)
    point_in = list(ashape.centroid.coords)[0]
    point_out = (100, 100)

    assert geometry.contains(chull, *point_in)
    assert geometry.contains(ashape, *point_in)
    assert geometry.contains(ashape, *point_in)
    assert geometry.contains(bbox, *point_in)

    assert not (geometry.contains(chull, *point_out))
    assert not (geometry.contains(ashape, *point_out))
    assert not (geometry.contains(ashape, *point_out))
    assert not (geometry.contains(bbox, *point_out))

    numpy.testing.assert_array_equal(bbox, geometry.bbox(bbox))
    numpy.testing.assert_array_equal(bbox, geometry.bbox(ashape))
    numpy.testing.assert_array_equal(bbox, geometry.bbox(ashape))
    numpy.testing.assert_array_equal(bbox, geometry.bbox(chull))
    numpy.testing.assert_array_equal(bbox, geometry.bbox(points))


def test_tree_functions():
    kdtree = ripley._build_best_tree(points, "euclidean")
    balltree = ripley._build_best_tree(points, "haversine")
    try:
        failtree = ripley._build_best_tree(points, "notametric")
    except KeyError:
        pass
    except:
        raise AssertionError("Failed to raise an error for _build_best_tree")

    with pytest.warns(UserWarning):
        mytree = ripley._build_best_tree(points, lambda u, v: numpy.var(u - v))

    # check that neighbors are not returned as a self-neighbor
    # for self-neighbor queries
    distances, indices = ripley._k_neighbors(kdtree, points, k=1)
    assert (indices.squeeze() != numpy.arange(points.shape[0])).all()
    distances, indices = ripley._k_neighbors(balltree, points, k=1)
    assert (indices.squeeze() != numpy.arange(points.shape[0])).all()
    distances, indices = ripley._k_neighbors(mytree, points, k=1)
    assert (indices.squeeze() != numpy.arange(points.shape[0])).all()


def test_prepare():
    tmp_bbox = ripley._prepare_hull(points, "bbox")
    numpy.testing.assert_array_equal(bbox, tmp_bbox)

    tmp_bbox = ripley._prepare_hull(points, None)
    numpy.testing.assert_array_equal(bbox, tmp_bbox)

    tmp_bbox = ripley._prepare_hull(points, bbox)
    assert tmp_bbox is bbox  # pass-through with no modification

    tmp_ashape = ripley._prepare_hull(points, "alpha")
    assert tmp_ashape.equals(ashape)

    tmp_ashape = ripley._prepare_hull(points, "α")
    assert tmp_ashape.equals(ashape)

    tmp_ashape = ripley._prepare_hull(points, ashape)
    assert tmp_ashape is ashape  # pass-through with no modification

    tmp_ashape = ripley._prepare_hull(points, ashape)
    assert shapely.equals(tmp_ashape, ashape)

    tmp_chull = ripley._prepare_hull(points, chull)
    assert tmp_chull is chull  # pass-through with no modification

    tmp_chull = ripley._prepare_hull(points, "convex")
    numpy.testing.assert_allclose(tmp_chull.equations, chull.equations)

    # --------------------------------------------------------------------------
    # Now, check the prepare generally
    # check edge correction raise

    try:
        ripley._prepare(points, None, None, "euclidean", ashape, "ripley")
        raise AssertionError()
    except NotImplementedError:
        pass
    except AssertionError:
        raise AssertionError("Did not raise an error when edge correction is set")

    # check tree gets converted into data with no tree
    out = ripley._prepare(tree, None, None, "euclidean", ashape, None)
    numpy.testing.assert_array_equal(points, out[0])

    # check three distance metrics
    out = ripley._prepare(tree, None, None, "euclidean", ashape, None)[3]
    assert out == "euclidean"
    out = ripley._prepare(tree, None, None, "haversine", ashape, None)[3]
    assert out == "haversine"
    test_func = lambda u, v: numpy.var(u - v)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        out = ripley._prepare(tree, None, None, test_func, ashape, None)[3]
        assert out is test_func

    # check precomputed failure
    try:
        out = ripley._prepare(tree, None, None, "precomputed", ashape, None)
        raise AssertionError()
    except ValueError:
        pass
    except AssertionError:
        raise AssertionError(
            'Did not raise when metric="precomputed" but' " no distances provided"
        )

    # check support setting will:
    # give 20 breaks from 0 to max dist if none
    out = ripley._prepare(tree, None, None, "euclidean", ashape, None)[1]
    assert len(out) == 20
    assert out.min() == 0
    numpy.testing.assert_allclose(out.max(), 34.631242)
    numpy.testing.assert_allclose(out.min(), 0)
    out = ripley._prepare(tree, 30, None, "euclidean", ashape, None)[1]
    assert len(out) == 30
    numpy.testing.assert_allclose(out.max(), 34.631242)
    numpy.testing.assert_allclose(out.min(), 0)
    # give tuple correctly for 1, 2, and 3-length tuples
    out = ripley._prepare(tree, (4,), None, "euclidean", ashape, None)[1]
    assert out.max() == 4
    out = ripley._prepare(tree, (2, 10), None, "euclidean", ashape, None)[1]
    assert out.max() == 10
    assert out.min() == 2
    out = ripley._prepare(tree, (2, 10, 5), None, "euclidean", ashape, None)[1]
    assert out.max() == 10
    assert out.min() == 2
    assert len(out) == 5
    # passthrough support
    out = ripley._prepare(tree, numpy.arange(40), None, "euclidean", ashape, None)[1]
    assert len(out) == 40
    assert (out == numpy.arange(40)).all()


def test_simulate():
    assert random.poisson(ashape).shape == (100, 2)
    assert random.poisson(chull).shape == (100, 2)
    assert random.poisson(bbox).shape == (100, 2)

    assert random.poisson(ashape, intensity=1e-2).shape == (50, 2)
    assert random.poisson(chull, intensity=1e-2).shape == (52, 2)
    assert random.poisson(bbox, intensity=1e-2).shape == (76, 2)

    assert random.poisson(ashape, size=90).shape == (90, 2)
    assert random.poisson(chull, intensity=1e-2).shape == (52, 2)
    assert random.poisson(bbox, intensity=1e-2, size=3).shape == (3, 76, 2)
    assert random.poisson(bbox, intensity=None, size=(10, 4)).shape == (4, 10, 2)

    # still need to check the other simulators
    # normal
    # cluster poisson
    # cluster normal


def test_f():
    # -------------------------------------------------------------------------#
    # Check f function has consistent performance

    nn_other = D_other.min(axis=0)
    n_obs_at_dist, histogram_support = numpy.histogram(nn_other, bins=support)
    manual_f = numpy.asarray([0, *numpy.cumsum(n_obs_at_dist) / n_obs_at_dist.sum()])
    numpy.random.seed(2478879)
    f_test = ripley.f_test(points, support=support, distances=D_other, n_simulations=99)

    numpy.testing.assert_allclose(support, f_test.support)
    numpy.testing.assert_allclose(manual_f, f_test.statistic)
    numpy.testing.assert_allclose(
        f_test.pvalue < 0.05, [1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1]
    )
    assert f_test.simulations is None

    f_test = ripley.f_test(
        points,
        support=support,
        distances=D_other,
        n_simulations=99,
        keep_simulations=True,
    )
    assert f_test.simulations.shape == (99, 15)


def test_g():
    # -------------------------------------------------------------------------#
    # Check f function works, has statistical results that are consistent

    nn_self = (D_self + numpy.eye(points.shape[0]) * 10000).min(axis=0)
    n_obs_at_dist, histogram_support = numpy.histogram(nn_self, bins=support)
    numpy.random.seed(2478879)
    manual_g = numpy.asarray([0, *numpy.cumsum(n_obs_at_dist) / n_obs_at_dist.sum()])
    g_test = ripley.g_test(points, support=support, n_simulations=99)

    numpy.testing.assert_allclose(support, g_test.support)
    numpy.testing.assert_allclose(manual_g, g_test.statistic)
    numpy.testing.assert_allclose(
        g_test.pvalue < 0.05, [1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    )
    assert g_test.simulations is None

    g_test = ripley.g_test(
        points, support=support, n_simulations=99, keep_simulations=True
    )
    assert g_test.simulations.shape == (99, 15)


def test_j():
    # -------------------------------------------------------------------------#
    # Check j function works, matches manual, is truncated correctly

    numpy.random.seed(2478879)
    j_test = ripley.j_test(points, support=support, n_simulations=99, truncate=True)
    numpy.random.seed(2478879)
    j_test_fullno = ripley.j_test(
        points, support=support, n_simulations=0, truncate=False
    )
    numpy.testing.assert_array_equal(j_test.support[:4], support[:4])
    numpy.testing.assert_array_equal(j_test_fullno.support, support)
    numpy.random.seed(2478879)
    _, f_tmp = ripley.f(points, support=support)
    _, g_tmp = ripley.g(points, support=support)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        manual_j = (1 - g_tmp) / (1 - f_tmp)
    assert numpy.isnan(manual_j[-1])
    assert len(manual_j) > len(j_test.support)
    assert len(manual_j) == len(j_test_fullno.support)

    numpy.testing.assert_allclose(j_test.statistic, manual_j[:4], atol=0.1, rtol=0.05)


def test_k():
    # -------------------------------------------------------------------------#
    # Check K function works, matches a manual, slower explicit computation

    k_test = ripley.k_test(points, support=support)
    n = points.shape[0]
    intensity = n / ripley._area(bbox)
    manual_unscaled_k = numpy.asarray(
        [(d_self < d).sum() for d in support], dtype=float
    )
    numpy.testing.assert_allclose(
        k_test.statistic, manual_unscaled_k * 2 / n / intensity
    )


def test_l():
    # -------------------------------------------------------------------------#
    # Check L Function works, can be linearized, and has the right value
    _, k = ripley.k(points, support=support)
    l_test = ripley.l_test(points, support=support, n_simulations=0)
    l_test_lin = ripley.l_test(
        points, support=support, n_simulations=0, linearized=True
    )

    numpy.testing.assert_allclose(l_test.statistic, numpy.sqrt(k / numpy.pi))
    numpy.testing.assert_allclose(
        l_test_lin.statistic, numpy.sqrt(k / numpy.pi) - l_test.support
    )



"""
Simulation of planar point processes

TODO

- inhibition process(es)
- optimize draws for complex windows
- documentation
"""

__author__ = "Serge Rey sjsrey@gmail.com"
__all__ = ["PointProcess", "PoissonPointProcess", "PoissonClusterPointProcess"]

import numpy as np
import libpysal as ps
from numpy.random import poisson
from .pointpattern import PointPattern as PP
import warnings

warnings.filterwarnings(
    "ignore", "Objects based on the `Geometry` class will", FutureWarning
)


def runif_in_circle(n, radius=1.0, center=(0.0, 0.0), burn=2, verbose=False):
    """
    Generate n points within a circle of given radius.

    Parameters
    ----------
    n             : int
                    Number of points.
    radius        : float
                    Radius of the circle.
    center        : tuple
                    Coordinates of the center.

    Returns
    -------
                  : array
                    (n+1, 2), coordinates of generated points as well as
                    the center.

    """

    good = np.zeros((n, 2), float)
    c = 0
    r = radius
    r2 = r * r
    it = 0
    while c < n:
        x = np.random.uniform(-r, r, (burn * n, 1))
        y = np.random.uniform(-r, r, (burn * n, 1))
        ids = np.where(x * x + y * y <= r2)
        candidates = np.hstack((x, y))[ids[0]]
        nc = candidates.shape[0]
        need = n - c
        if nc > need:  # more than we need
            good[c:] = candidates[:need]
        else:  # use them all and keep going
            good[c : c + nc] = candidates
        c += nc
        it += 1
    if verbose:
        print("Iterations: {}".format(it))
    return good + np.asarray(center)


class PointProcess(object):
    """
    Point Process base class.

    Parameters
    ----------
    window        : :py:class:`~.window.Window`
                    Bounding geometric object to contain point process
                    realizations.
    n             : int
                    Size of each realization.
    samples       : list
                    Number of realizations.
    asPP          : bool
                    Control the data type of value in the "realizations"
                    dictionary. If True, the data type is point
                    pattern as defined in pointpattern.py; if False,
                    the data type is an two-dimensional array.

    Attributes
    ----------
    realizations  : dictionary
                    The key is the index of each realization, and the
                    value is simulated event points for each
                    realization. The data type of the value is
                    controlled by the parameter "asPP".
    parameters    : dictionary
                    Dictionary of a dictionary.
                    The key is the index of each realization, and the
                    value is a dictionary with the key 'n' and the
                    value size of each realization.

    """

    def __init__(self, window, n, samples, asPP=False, **args):
        super(PointProcess, self).__init__()
        self.window = window
        self.n = n
        self.samples = samples
        self.args = args
        self.realizations = {}
        self.setup()
        for sample in range(samples):
            self.realizations[sample] = self.draw(self.parameters[sample])
        if asPP:
            for sample in self.realizations:
                points = self.realizations[sample]
                self.realizations[sample] = PP(points, window=self.window)
        warnings.warn(
            "These point pattern simulators are deprecated! Please replace them"
            " with equivalent functions in pointpats.random",
            DeprecationWarning,
            stacklevel=2,
        )

    def draw(self, parameter):
        """
        Generate a series of point coordinates within the given window.

        Parameters
        ----------
        parameter  : dictionary
                     Key: 'n'.
                     Value: size of the realization.

        Returns
        -------
                   : array
                     A series of point coordinates.

        """
        c = 0
        sample = []
        n = parameter["n"]
        while c < n:
            pnts = self.realize(n)
            pnts = [ps.cg.shapes.Point((x, y)) for x, y in pnts]
            pins = self.window.filter_contained(pnts)
            sample.extend(pins)
            c = len(sample)
        return np.array([np.asarray(p) for p in sample[:n]])

    def realize(self):
        pass

    def setup(self):
        pass


class PoissonPointProcess(PointProcess):
    """
    Poisson point process including :math:`N`-conditioned CSR process and
    :math:`\lambda`-conditioned CSR process.

    Parameters
    ----------
    window        : :py:class:`~.window.Window`
                    Bounding geometric object to contain point process
                    realizations.
    n             : int
                    Size of each realization.
    samples       : list
                    Number of realizations.
    conditioning  : bool
                    If True, use the :math:`\lambda`-conditioned CSR process,
                    number of events would vary across realizations;
                    if False, use the :math:`N`-conditioned CSR process.
    asPP          : bool
                    Control the data type of value in the "realizations"
                    dictionary. If True, the data type is point
                    pattern as defined in pointpattern.py; if False,
                    the data type is an two-dimensional array.

    Attributes
    ----------
    realizations  : dictionary
                    The key is the index of each realization, and the
                    value is simulated event points for each
                    realization. The data type of the value is
                    controlled by the parameter "asPP".
    parameters    : dictionary
                    Dictionary of a dictionary.
                    The key is the index of each realization, and the
                    value is a dictionary with the key 'n' and the
                    value:
                    1. always equal to the parameter n in the case of
                    N-conditioned process.
                    For example, {0:{'n':100},1:{'n':100},2:{'n':100}}
                    2. randomly generated from a Possion process in
                    the case of lambda-conditioned process.
                    For example, {0:{'n':97},1:{'n':100},2:{'n':98}}

    Examples
    --------
    >>> import libpysal as ps
    >>> import numpy as np
    >>> from pointpats import Window
    >>> from libpysal.cg import shapely_ext

    Open the virginia polygon shapefile

    >>> va = ps.io.open(ps.examples.get_path("virginia.shp"))

    Create the exterior polygons for VA from the union of the county shapes

    >>> polys = [shp for shp in va]
    >>> state = shapely_ext.cascaded_union(polys)

    Create window from virginia state boundary

    >>> window = Window(state.parts)

    1. Simulate a :math:`N`-conditioned csr process in the same window (10
    points, 2 realizations)

    >>> np.random.seed(5)
    >>> samples1 = PoissonPointProcess(window, 10, 2, conditioning=False, asPP=False)
    >>> samples1.realizations[0] # the first realized event points
    array([[-81.80326547,  36.77687577],
           [-78.5166233 ,  37.34055832],
           [-77.21660795,  37.7491503 ],
           [-79.30361037,  37.40467853],
           [-78.61625258,  36.61234487],
           [-81.43369537,  37.13784646],
           [-80.91302108,  36.60834063],
           [-76.90806444,  37.95525903],
           [-76.33475868,  36.62635347],
           [-79.71621808,  37.27396618]])

    2. Simulate a :math:`\lambda`-conditioned csr process in the same window (10
    points, 2 realizations)

    >>> np.random.seed(5)
    >>> samples2 = PoissonPointProcess(window, 10, 2, conditioning=True, asPP=True)
    >>> samples2.realizations[0].n # the size of first realized point pattern
    10
    >>> samples2.realizations[1].n # the size of second realized point pattern
    13

    """

    def __init__(self, window, n, samples, conditioning=False, asPP=False):
        self.conditioning = conditioning
        super(PoissonPointProcess, self).__init__(window, n, samples, asPP)

    def setup(self):
        """
        Generate the number of events for each realization. If
        "conditioning" is False, all the event numbers are the same;
        if it is True, the event number is a random variable
        following a Poisson distribution.

        """

        self.parameters = {}
        if self.conditioning:
            lambdas = poisson(self.n, self.samples)
            for i, l in enumerate(lambdas):
                self.parameters[i] = {"n": l}
        else:
            for i in range(self.samples):
                self.parameters[i] = {"n": self.n}

    def realize(self, n):
        """
        Generate n points which are randomly and independently
        distributed in the minimum bounding box of "window".

        Parameters
        ----------
        n             : int
                        Number of point events.

        Returns
        -------
                      : array
                        (n,2), n point coordinates.

        """

        l, b, r, t = self.window.bbox
        xs = np.random.uniform(l, r, (n, 1))
        ys = np.random.uniform(b, t, (n, 1))
        return zip(xs, ys)


class PoissonClusterPointProcess(PointProcess):
    """
    Poisson cluster point process (Neyman Scott).
    Two stages:
    1. parent CSR process: :math:`N`-conditioned or
    :math:`\lambda`-conditioned. If parent events follow a
    :math:`\lambda`-conditioned CSR process,
    the number of parent events varies across realizations.
    2. child process: fixed number of points in circle centered
    on each parent.

    Parameters
    ----------
    window        : :py:class:`~.window.Window`
                    Bounding geometric object to contain point process
                    realizations.
    n             : int
                    Size of each realization.
    parents       : int
                    Number of parents.
    radius        : float
                    Radius of the circle centered on each parent.
    samples       : list
                    Number of realizations.
    asPP          : bool
                    Control the data type of value in the "realizations"
                    dictionary. If True, the data type is point
                    pattern as defined in pointpattern.py; if False,
                    the data type is an two-dimensional array.
    conditioning  : bool
                    If True, use the :math:`lambda`-conditioned CSR process
                    for parent events, leading to varied number of
                    parent events across realizations;
                    if False, use the :math:`N`-conditioned CSR process.

    Attributes
    ----------
    children      : int
                    Number of childrens centered on each parent. Can
                    be considered as local intensity.
    num_parents   : dictionary
                    The key is the index of each realization. The
                    value is the number of parent events for each
                    realization.
    realizations  : dictionary
                    The key is the index of each realization, and the
                    value is simulated event points for each
                    realization. The data type of the value is
                    controlled by the parameter "asPP".
    parameters    : dictionary
                    Dictionary of a dictionary.
                    The key is the index of each realization, and the
                    value is a dictionary with the key 'n' and the
                    value always equal to the parameter n in the
                    case of
                    N-conditioned process.
                    For example, {0:{'n':100},1:{'n':100},2:{'n':100}}
                    2. randomly generated from a Possion process in
                    the case of lambda-conditioned process.
                    For example, {0:{'n':97},1:{'n':100},2:{'n':98}}

    Examples
    --------
    >>> import libpysal as ps
    >>> import numpy as np
    >>> from pointpats import Window
    >>> from libpysal.cg import shapely_ext

    Open the virginia polygon shapefile

    >>> va = ps.io.open(ps.examples.get_path("virginia.shp"))

    Create the exterior polygons for VA from the union of the county shapes

    >>> polys = [shp for shp in va]
    >>> state = shapely_ext.cascaded_union(polys)

    Create window from virginia state boundary

    >>> window = Window(state.parts)

    1. Simulate a Poisson cluster process of size 200 with 10 parents
    and 20 children within 0.5 units of each parent
    (parent events:  :math:`N`-conditioned CSR)

    >>> np.random.seed(10)
    >>> samples1 = PoissonClusterPointProcess(window, 200, 10, 0.5, 1, asPP=True, conditioning=False)
    >>> samples1.parameters # number of events for the realization
    {0: {'n': 200}}
    >>> samples1.num_parents #number of parent events for each realization
    {0: 10}
    >>> samples1.children # number of children events centered on each parent event
    20

    2. Simulate a Poisson cluster process of size 200 with 10 parents
    and 20 children within 0.5 units of each parent
    (parent events:  :math:`\lambda`-conditioned CSR)

    >>> np.random.seed(10)
    >>> samples2 = PoissonClusterPointProcess(window, 200, 10, 0.5, 1, asPP=True, conditioning=True)
    >>> samples2.parameters # number of events for the realization might not be equal to 200
    {0: {'n': 260}}
    >>> samples2.num_parents #number of parent events for each realization
    {0: 13}
    >>> samples2.children # number of children events centered on each parent event
    20

    """

    def __init__(
        self,
        window,
        n,
        parents,
        radius,
        samples,
        keep=False,
        asPP=False,
        conditioning=False,
    ):
        self.conditioning = conditioning
        self.parents = parents
        self.children = int(np.ceil(n * 1.0 / parents))
        self.radius = radius
        self.keep = keep
        super(PoissonClusterPointProcess, self).__init__(window, n, samples, asPP)

    def setup(self):
        """
        Generate the number of events for each realization. If
        "conditioning" is False, all the event numbers are the same;
        if it is True, the number of parents is a random variable
        following a Poisson distribution, resulting in varied number
        of events.

        """

        self.parameters = {}
        self.num_parents = {}
        if self.conditioning:
            lambdas = poisson(self.parents, self.samples)
            for i, l in enumerate(lambdas):
                num = l * self.children
                self.parameters[i] = {"n": num}
                self.num_parents[i] = l
        else:
            for i in range(self.samples):
                self.parameters[i] = {"n": self.n}
                self.num_parents[i] = self.parents

    def realize(self, n):
        """
        Generate n points which are distributed in a clustered
        fashion in the minimum bounding box of "window".

        Parameters
        ----------
        n             : int
                        Number of point events.

        Returns
        -------
        res           : array
                        (n,2), n point coordinates.

        """
        l, b, r, t = self.window.bbox
        d = self.radius
        # get parent points
        pxs = np.random.uniform(l, r, (int(n / self.children), 1))
        pys = np.random.uniform(b, t, (int(n / self.children), 1))
        cents = np.hstack((pxs, pys))
        # generate children points
        pnts = [runif_in_circle(self.children, d, center) for center in cents]
        res = np.vstack(np.asarray(pnts))
        if self.keep:
            res = np.vstack((np.asarray(cents), res))
        np.random.shuffle(res)  # so we don't truncate in a biased fashion
        return res



# TODO: skyum, dtot, weighted_mean_center, manhattan_median
import unittest
import numpy as np

from ..centrography import *

from libpysal.common import RTOL


class TestCentrography(unittest.TestCase):
    def setUp(self):
        self.points = np.array(
            [
                [66.22, 32.54],
                [22.52, 22.39],
                [31.01, 81.21],
                [9.47, 31.02],
                [30.78, 60.10],
                [75.21, 58.93],
                [79.26, 7.68],
                [8.23, 39.93],
                [98.73, 77.17],
                [89.78, 42.53],
                [65.19, 92.08],
                [54.46, 8.48],
            ]
        )

    def test_centrography_mar(self):
        mrr = minimum_rotated_rectangle(self.points)
        known = np.array(
            [
                [36.40165, 104.61744],
                [4.087286, 30.417522],
                [75.59908, -0.726158],
                [107.913445, 73.47376251220703],
            ]
        )
        for i in range(5):
            success = np.allclose(mrr, np.roll(known, i, axis=0))
            if success:
                break
        if not success:
            raise AssertionError(
                f"Minimum Rotated Rectangle cannot be"
                f"aligned with correct answer:"
                f"\ncomputed {mrr}\nknown: {known}"
            )

    def test_centrography_mbr(self):
        min_x, min_y, max_x, max_y = minimum_bounding_rectangle(self.points)
        np.testing.assert_allclose(min_x, 8.2300000000000004, RTOL)
        np.testing.assert_allclose(min_y, 7.6799999999999997, RTOL)
        np.testing.assert_allclose(max_x, 98.730000000000004, RTOL)
        np.testing.assert_allclose(max_y, 92.079999999999998, RTOL)

    def test_centrography_hull(self):
        hull_array = hull(self.points)
        res = np.array(
            [
                [31.01, 81.21],
                [8.23, 39.93],
                [9.47, 31.02],
                [22.52, 22.39],
                [54.46, 8.48],
                [79.26, 7.68],
                [89.78, 42.53],
                [98.73, 77.17],
                [65.19, 92.08],
            ]
        )
        np.testing.assert_array_equal(hull_array, res)

    def test_centrography_mean_center(self):
        res = np.array([52.57166667, 46.17166667])
        np.testing.assert_array_almost_equal(mean_center(self.points), res)

    def test_centrography_std_distance(self):
        std = std_distance(self.points)
        np.testing.assert_allclose(std, 40.149806489086714, RTOL)

    def test_centrography_ellipse(self):
        res = ellipse(self.points)
        np.testing.assert_allclose(res[0], 39.623867886462982, RTOL)
        np.testing.assert_allclose(res[1], 42.753818949026815, RTOL)
        np.testing.assert_allclose(res[2], 1.1039268428650906, RTOL)

    def test_centrography_euclidean_median(self):
        euclidean = euclidean_median(self.points)
        res = np.array([54.16770671, 44.4242589])
        np.testing.assert_array_almost_equal(euclidean, res, decimal=3)



import numpy as np


def plot_density(
    data,
    bandwidth,
    kernel=None,
    resolution=100,
    levels=10,
    fill=False,
    margin=0.1,
    ax=None,
    figsize=None,
    **kwargs,
):
    """Plot kernel density of a given point pattern

    The KDE can be done either using :class:`statsmodels.nonparametric.KDEMultivariate`,
    which is used when ``kernel=None``, or using :class:`KDEpy.FFTKDE` when kernel is
    set. :class:`~KDEpy.FFTKDE` tends to be generally faster in most cases but may need
    different than ``"gaussian"`` kernel to resolve in some cases. For small data of up
    to 10 000 points, the difference is not noticeable. For larger data, specify
    ``bandwidth`` to enforce the use of :class:`~KDEpy.FFTKDE`. Note that while being
    faster, :class:`~KDEpy.FFTKDE` may in some case result in erroneous KDE.

    KDE is plotted using matplotlib's :meth:`~matplotlib.pyplot.contour` or
    :meth:`~matplotlib.pyplot.contourf` function to plot the density.

    If MultiPoints are given, each point is treated as separate observation.

    Parameters
    ----------
    data : array or geopandas object
        Array with a shape (2, n) containing coordinates of points
        or a geopandas object with (Multi)Point geometry. Assumes
        projected coordinates, geographical coordinates (latitude, longitude)
        are not supported.
    bandwidth : float
        bandwidth in the units of CRS in which data is
    kernel : str | None, optional
        The kernel function. If None, defaults to the Gaussian kernel and statsmodels
        implementation. If set, uses KDEpy implementation. See
        :meth:`KDEpy.FFTKDE._available_kernels.keys()` for choices.
    resolution : int | tuple(int, int), optional
        resolution of the grid used to evaluate the probability density
        function. If tuple, each dimension of the grid is specified separately.
        By default 100
    levels : int or array-like, optional
        Determines the number and positions of the contour lines / regions.
        See the documentation of :meth:`~matplotlib.pyplot.contour` for details.
        By default 10
    fill : bool, optional
        Fill the area between contour lines, by default False
    margin : float, optional
        The factor of the margin by which the extent of the data will be expanded when
        creating the grid. 0.1 means 10% on each side, by default 0.1. Only used
        with the ``statsmodels`` implementation.
    ax : matplotlib.axes.Axes (default None)
        axes on which to draw the plot
    figsize : tuple of integers (default None)
        Size of the resulting ``matplotlib.figure.Figure``. If the argument
        ``ax`` is given explicitly, ``figsize`` is ignored.
    **kwargs
        Keyword arguments passed to :meth:`~matplotlib.pyplot.contour` or
        :meth:`~matplotlib.pyplot.contourf` used for further
        styling of the plot, for example ``cmap``, ``linewidths``, ``linestyles``,
        or `alpha`. See the documentation of :meth:`~matplotlib.pyplot.contour` for
        details.

    Returns
    -------
    matplotlib.axes.Axes
        matplotlib axes instance with the contour plot
    """
    if kernel is None:
        try:
            import statsmodels.api as sm
        except ImportError as err:
            raise ImportError(
                "statsmodels is required for `plot_density` when kernel"
                "is not specified."
            ) from err

        engine = "sm"
    else:
        try:
            from KDEpy import FFTKDE
        except ImportError as err:
            raise ImportError(
                "KDEpy is required for `plot_density` when kernel is not None."
            ) from err

        engine = "kdepy"

    try:
        import matplotlib.pyplot as plt
    except ImportError as err:
        raise ImportError("matplotlib is required for `plot_density`") from err

    if ax is None:
        _, ax = plt.subplots(figsize=figsize)

    ax.set_aspect("equal") # bandwidth is fixed, hence aspect shall be equal

    if isinstance(data, np.ndarray):
        pass
    else:  # geopandas
        if not data.geom_type.str.contains("Point").all():
            raise ValueError(
                "data contain non-point geometries. "
                "Only (Multi)Points are supported."
            )
        data = data.get_coordinates().values

    if engine == "sm":
        dens_u = sm.nonparametric.KDEMultivariate(
            data=[data[:, 0], data[:, 1]],
            var_type="cc",
            bw=[bandwidth, bandwidth],
        )

        xmax = data[:, 0].max()
        xmin = data[:, 0].min()
        ymax = data[:, 1].max()
        ymin = data[:, 1].min()

        # get margin to go beyond the extent to avoid cutting of countour lines
        x_margin = (xmax - xmin) * margin
        y_margin = (ymax - ymin) * margin

        if isinstance(resolution, tuple):
            x_res, y_res = resolution
        elif isinstance(resolution, (float, int)):
            x_res = resolution
            y_res = resolution
        elif resolution is None:
            x_res = 100
            y_res = 100
        else:
            raise ValueError("Unsupported option for `resolution`.")

        # create mesh for predicting KDE on with more space around the points
        x_mesh, y_mesh = np.meshgrid(
            np.linspace(xmin - x_margin, xmax + x_margin, x_res),
            np.linspace(ymin - y_margin, ymax + y_margin, y_res),
        )

        # get the prediction
        pred = dens_u.pdf(np.vstack([x_mesh.flatten(), y_mesh.flatten()]).T)
        z = pred.reshape(x_mesh.shape)

    else:
        kde = FFTKDE(bw=bandwidth, kernel=kernel)
        grid, points = kde.fit(data).evaluate(resolution)
        x_mesh, y_mesh = np.unique(grid[:, 0]), np.unique(grid[:, 1])
        z = points.reshape(resolution, resolution).T

    if fill:
        ax.contourf(x_mesh, y_mesh, z, levels=levels, **kwargs)
    else:
        ax.contour(x_mesh, y_mesh, z, levels=levels, **kwargs)

    return ax



from warnings import warn

import geopandas as gpd
import libpysal as lps
import numpy
from pytest import approx
import pytest
import matplotlib.pyplot as plt

from pointpats import (
    Knox,
    KnoxLocal,
    SpaceTimeEvents,
    jacquez,
    knox,
    mantel,
    modified_knox,
)


class TestKnox:
    def setup_method(self):
        path = lps.examples.get_path("burkitt.shp")
        self.gdf = gpd.read_file(path)

    def test_knox(self):
        global_knox = Knox(self.gdf[["X", "Y"]], self.gdf[["T"]], delta=20, tau=5)
        assert global_knox.statistic_ == 13
        assert global_knox.p_poisson == 0.14624558197140414
        assert hasattr(global_knox, "sim") == False
        numpy.testing.assert_array_equal(
            global_knox.observed, [[1.300e01, 3.416e03], [3.900e01, 1.411e04]]
        )
        numpy.testing.assert_allclose(
            global_knox.expected,
            [[1.01438161e01, 3.41885618e03], [4.1856139e01, 1.41071438e04]],
            rtol=1e-5,
            atol=0,
        )

        numpy.random.seed(12345)
        global_knox = Knox(
            self.gdf[["X", "Y"]], self.gdf[["T"]], delta=20, tau=5, keep=True
        )
        assert global_knox.statistic_ == 13
        assert hasattr(global_knox, "sim") == True
        assert global_knox.p_sim == 0.21

    def test_knox_from_gdf(self):
        gdf = self.gdf.copy()
        # not technically the correct CRS...
        gdf.crs = 21096
        global_knox = Knox.from_dataframe(gdf, time_col="T", delta=20, tau=5)
        assert global_knox.statistic_ == 13
        assert global_knox.p_poisson == 0.14624558197140414
        assert hasattr(global_knox, "sim") == False
        numpy.testing.assert_array_equal(
            global_knox.observed, [[1.300e01, 3.416e03], [3.900e01, 1.411e04]]
        )
        numpy.testing.assert_allclose(
            global_knox.expected,
            [[1.01438161e01, 3.41885618e03], [4.1856139e01, 1.41071438e04]],
            rtol=1e-5,
            atol=0,
        )

        # no CRS should raise a warning
        global_knox = Knox.from_dataframe(self.gdf, time_col="T", delta=20, tau=5)
        numpy.testing.assert_allclose(
            global_knox.expected,
            [[1.01438161e01, 3.41885618e03], [4.1856139e01, 1.41071438e04]],
            rtol=1e-5,
            atol=0,
        )

        # unprojected coords
        try:
            gdf.crs = 4326
            global_knox = Knox.from_dataframe(gdf, time_col="T", delta=20, tau=5)
        except ValueError:
            warn("successfully caught crs error")
            pass

        # non-numeric type for time
        try:
            gdf["T"] = gdf["T"].astype("O")
            global_knox = Knox.from_dataframe(gdf, time_col="T", delta=20, tau=5)
        except ValueError:
            warn("successfully caught dtype error")
            pass

        numpy.random.seed(12345)
        global_knox = Knox(
            self.gdf[["X", "Y"]], self.gdf[["T"]], delta=20, tau=5, keep=True
        )
        assert global_knox.statistic_ == 13
        assert hasattr(global_knox, "sim") == True
        assert global_knox.p_sim == 0.21


class TestKnoxLocal:
    def setup_method(self):
        path = lps.examples.get_path("burkitt.shp")
        self.gdf = gpd.read_file(path)

    def test_knox_local(self):
        numpy.random.seed(12345)
        local_knox = KnoxLocal(
            self.gdf[["X", "Y"]].values,
            self.gdf[["T"]].values,
            delta=20,
            tau=5,
            keep=True,
        )
        assert local_knox.statistic_.shape == (188,)
        lres = local_knox
        gt0ids = numpy.where(lres.nsti > 0)
        numpy.testing.assert_array_equal(
            gt0ids,
            [
                [
                    25,
                    26,
                    30,
                    31,
                    35,
                    36,
                    41,
                    42,
                    46,
                    47,
                    51,
                    52,
                    102,
                    103,
                    116,
                    118,
                    122,
                    123,
                    137,
                    138,
                    139,
                    140,
                    158,
                    159,
                    162,
                    163,
                ]
            ],
        )
        numpy.testing.assert_allclose(
            lres.p_hypergeom[gt0ids],
            [
                0.1348993,
                0.14220663,
                0.07335085,
                0.08400282,
                0.1494317,
                0.21524073,
                0.0175806,
                0.04599869,
                0.17523687,
                0.18209188,
                0.19111321,
                0.16830444,
                0.13734428,
                0.14703242,
                0.06796364,
                0.03192559,
                0.13734428,
                0.17523687,
                0.12998154,
                0.1933476,
                0.13244507,
                0.13244507,
                0.12502644,
                0.14703242,
                0.12502644,
                0.12998154,
            ],
            rtol=1e-5,
            atol=0,
        )
        numpy.testing.assert_array_equal(
            lres.p_sims[gt0ids],
            [
                0.3,
                0.33,
                0.11,
                0.17,
                0.3,
                0.42,
                0.06,
                0.06,
                0.33,
                0.34,
                0.36,
                0.38,
                0.3,
                0.29,
                0.41,
                0.19,
                0.31,
                0.39,
                0.18,
                0.39,
                0.48,
                0.41,
                0.22,
                0.41,
                0.39,
                0.32,
            ],
        )

    def test_knox_local_from_gdf(self):
        gdf = self.gdf
        gdf.crs = 21096
        numpy.random.seed(12345)
        local_knox = KnoxLocal.from_dataframe(
            gdf, time_col="T", delta=20, tau=5, keep=True
        )
        assert local_knox.statistic_.shape == (188,)
        lres = local_knox
        gt0ids = numpy.where(lres.nsti > 0)
        numpy.testing.assert_array_equal(
            gt0ids,
            [
                [
                    25,
                    26,
                    30,
                    31,
                    35,
                    36,
                    41,
                    42,
                    46,
                    47,
                    51,
                    52,
                    102,
                    103,
                    116,
                    118,
                    122,
                    123,
                    137,
                    138,
                    139,
                    140,
                    158,
                    159,
                    162,
                    163,
                ]
            ],
        )
        numpy.testing.assert_allclose(
            lres.p_hypergeom[gt0ids],
            [
                0.1348993,
                0.14220663,
                0.07335085,
                0.08400282,
                0.1494317,
                0.21524073,
                0.0175806,
                0.04599869,
                0.17523687,
                0.18209188,
                0.19111321,
                0.16830444,
                0.13734428,
                0.14703242,
                0.06796364,
                0.03192559,
                0.13734428,
                0.17523687,
                0.12998154,
                0.1933476,
                0.13244507,
                0.13244507,
                0.12502644,
                0.14703242,
                0.12502644,
                0.12998154,
            ],
            rtol=1e-5,
            atol=0,
        )
        numpy.testing.assert_array_equal(
            lres.p_sims[gt0ids],
            [
                0.3,
                0.33,
                0.11,
                0.17,
                0.3,
                0.42,
                0.06,
                0.06,
                0.33,
                0.34,
                0.36,
                0.38,
                0.3,
                0.29,
                0.41,
                0.19,
                0.31,
                0.39,
                0.18,
                0.39,
                0.48,
                0.41,
                0.22,
                0.41,
                0.39,
                0.32,
            ],
        )

    def test_explore(self):
        gdf = self.gdf
        gdf.crs = 21096
        numpy.random.seed(12345)
        m = KnoxLocal.from_dataframe(
            gdf, time_col="T", delta=20, tau=5, keep=True
        ).explore()
        numpy.testing.assert_array_equal(
            m.get_bounds(),
            [
                [-0.0005034046601185694, 28.514258651567],
                [0.0008675512091255166, 28.514975377735905],
            ],
        )
        assert len(m.to_dict()["children"]) == 5

    @pytest.mark.mpl_image_compare
    def test_plot(self):
        gdf = self.gdf
        gdf.crs = 21096
        fig, ax2 = plt.subplots(figsize=(30,18))
        lk = KnoxLocal.from_dataframe(
            gdf, time_col="T", delta=20, tau=5, keep=True)
        lk.plot(inference='analytic', ax=ax2)
        return fig



# old tests refactored to pytest


class TestSpaceTimeEvents:
    def setup_method(self):
        path = lps.examples.get_path("burkitt.shp")
        self.events = SpaceTimeEvents(path, "T")

    def test_space_time_events(self):
        assert self.events.n == 188

    def test_knox(self):
        result = knox(self.events.space, self.events.t, delta=20, tau=5, permutations=1)
        assert result["stat"] == 13.0

    def test_mantel(self):
        result = mantel(
            self.events.space,
            self.events.time,
            1,
            scon=0.0,
            spow=1.0,
            tcon=0.0,
            tpow=1.0,
        )

        assert result["stat"] == approx(0.014154, rel=1e-4)

    def test_jacquez(self):
        result = jacquez(self.events.space, self.events.t, k=3, permutations=1)

        assert result["stat"] == 12

    def test_modified_knox(self):
        result = modified_knox(
            self.events.space, self.events.t, delta=20, tau=5, permutations=1
        )

        assert result["stat"] == approx(2.810160, rel=1e-4)



import numpy as np
import pytest

from pointpats import plot_density

matplotlib = pytest.importorskip("matplotlib")
statsmodels = pytest.importorskip("statsmodels")
KDEpy = pytest.importorskip("KDEpy")


class TestDensity:
    def setup_method(self):
        self.points = np.array(
            [
                [66.22, 32.54],
                [22.52, 22.39],
                [31.01, 81.21],
                [9.47, 31.02],
                [30.78, 60.10],
                [75.21, 58.93],
                [79.26, 7.68],
                [8.23, 39.93],
                [98.73, 77.17],
                [89.78, 42.53],
                [65.19, 92.08],
                [54.46, 8.48],
            ]
        )

    def test_default(self):
        ax = plot_density(self.points, 10)
        contourset = ax.collections[0]
        assert len(contourset.collections) == 12
        assert not contourset.filled
        np.testing.assert_array_equal(contourset.get_linewidths(), np.array([1.5] * 12))
        np.testing.assert_array_equal(
            contourset.get_edgecolor()[5],
            np.array([0.143343, 0.522773, 0.556295, 1.0]),
        )

    def test_bandwidth(self):
        ax = plot_density(self.points, 1)
        contourset = ax.collections[0]
        assert len(contourset.collections) == 10

    def test_resolution(self):
        ax = plot_density(self.points, 10, resolution=200)
        contourset = ax.collections[0]
        collections = contourset.collections
        assert len(collections) == 12

    def test_margin(self):
        ax = plot_density(self.points, 10, margin=0.3)
        contourset = ax.collections[0]
        collections = contourset.collections
        assert len(collections) == 12

    def test_kdepy(self):
        ax = plot_density(self.points, 10, kernel="gaussian")
        contourset = ax.collections[0]
        collections = contourset.collections
        assert len(collections) == 12
        np.testing.assert_array_equal(contourset.get_linewidths(), np.array([1.5] * 12))

    def test_levels(self):
        ax = plot_density(self.points, 10, levels=5)
        contourset = ax.collections[0]
        collections = contourset.collections
        assert len(collections) == 7

    def test_fill(self):
        ax = plot_density(self.points, 10, fill=True)
        contourset = ax.collections[0]
        assert contourset.get_edgecolor().shape == (0, 4)
        assert contourset.filled
        np.testing.assert_array_equal(
            contourset.get_facecolor()[0],
            np.array([0.279566, 0.067836, 0.391917, 1.0]),
        )

    def test_geopandas(self):
        geopandas = pytest.importorskip("geopandas")

        gs = geopandas.GeoSeries.from_xy(*self.points.T)
        ax = plot_density(gs, 10)
        contourset = ax.collections[0]
        collections = contourset.collections
        assert len(collections) == 12
        np.testing.assert_array_equal(contourset.get_linewidths(), np.array([1.5] * 12))

    def test_kwargs(self):
        ax = plot_density(
            self.points, 10, cmap="magma", linewidths=0.5, linestyles="-."
        )
        contourset = ax.collections[0]
        collections = contourset.collections
        assert len(collections) == 12
        np.testing.assert_array_equal(contourset.get_linewidths(), np.array([0.5] * 12))

        np.testing.assert_array_equal(
            contourset.get_edgecolor()[5],
            np.array([0.639216, 0.189921, 0.49415, 1.0]),
        )



import numpy
import warnings
from scipy import spatial, interpolate
from collections import namedtuple
from .geometry import (
    area as _area,
    k_neighbors as _k_neighbors,
    build_best_tree as _build_best_tree,
    prepare_hull as _prepare_hull,
    TREE_TYPES,
)
from .random import poisson


__all__ = [
    "f",
    "g",
    "k",
    "j",
    "l",
    "f_test",
    "g_test",
    "k_test",
    "j_test",
    "l_test",
]


def _prepare(coordinates, support, distances, metric, hull, edge_correction):
    """
    prepare the arguments to convert into a standard format
    1. cast the coordinates to a numpy array
    2. precomputed metrics must have distances provided
    3. metrics must be callable or string
    4. warn if distances are specified and metric is not default
    5. make distances a numpy.ndarray
    6. construct the support, accepting:
        - num_steps -> a linspace with len(support) == num_steps
                       from zero to a quarter of the bounding box's smallest side
        - (stop, ) -> a linspace with len(support) == 20
                 from zero to stop
        - (start, stop) -> a linspace with len(support) == 20
                           from start to stop
        - (start, stop, num_steps) -> a linspace with len(support) == num_steps
                                      from start to stop
        - numpy.ndarray -> passed through
    """
    # Throw early if edge correction is requested
    if edge_correction is not None:
        raise NotImplementedError("Edge correction is not currently implemented.")

    # cast to coordinate array
    if isinstance(coordinates, TREE_TYPES):
        tree = coordinates
        coordinates = tree.data
    else:
        coordinates = numpy.asarray(coordinates)
    hull = _prepare_hull(coordinates, hull)

    # evaluate distances
    if (distances is None) and metric == "precomputed":
        raise ValueError(
            "If metric =`precomputed` then distances must"
            " be provided as a (n,n) numpy array."
        )
    if not (isinstance(metric, str) or callable(metric)):
        raise TypeError(
            f"`metric` argument must be callable or a string. Recieved: {metric}"
        )
    if distances is not None and metric != "euclidean":
        warnings.warn(
            "Distances were provided. The specified metric will be ignored."
            " To use precomputed distances with a custom distance metric,"
            " do not specify a `metric` argument.",
            stacklevel=2,
        )
        metric = "euclidean"

    if support is None:
        support = 20

    if isinstance(support, int):  # if just n_steps, use the max nnd
        # this is O(n log n) for kdtrees & balltrees
        tmp_tree = _build_best_tree(coordinates, metric=metric)
        max_dist = _k_neighbors(tmp_tree, coordinates, 1)[0].max()
        support = numpy.linspace(0, max_dist, num=support)
    # otherwise, we need to build it using (start, stop, step) semantics
    elif isinstance(support, tuple):
        if len(support) == 1:  # assuming this is with zero implicit start
            support = numpy.linspace(0, support[0], num=20)  # default support n bins
        elif len(support) == 2:
            support = numpy.linspace(*support, num=20)  # default support n bins
        elif len(support) == 3:
            support = numpy.linspace(support[0], support[1], num=support[2])
    else:  # try to use it as is
        try:
            support = numpy.asarray(support)
        except:
            raise TypeError(
                "`support` must be a tuple (either (start, stop, step), (start, stop) or (stop,)),"
                " an int describing the number of breaks to use to evalute the function,"
                " or an iterable containing the breaks to use to evaluate the function."
                " Recieved object of type {}: {}".format(type(support), support)
            )

    return coordinates, support, distances, metric, hull, edge_correction


# ------------------------------------------------------------#
# Statistical Functions                                       #
# ------------------------------------------------------------#


def f(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
):
    """
    Ripley's F function

    The so-called "empty space" function, this is the cumulative density function of
    the distances from a random set of points to the known points in the pattern.

    Parameters
    ----------
    coordinates : numpy.ndarray of shape (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.

    Returns
    -------
    a tuple containing the support values used to evalute the function
    and the values of the function at each distance value in the support.
    """
    coordinates, support, distances, metric, hull, _ = _prepare(
        coordinates, support, distances, metric, hull, edge_correction
    )
    if distances is not None:
        n = coordinates.shape[0]
        if distances.ndim == 2:
            k, p = distances.shape
            if k == p == n:
                warnings.warn(
                    f"A full distance matrix is not required for this function, and"
                    f" the intput matrix is a square {n},{n} matrix. Only the"
                    f" distances from p random points to their nearest neighbor within"
                    f" the pattern is required, as an {n},p matrix. Assuming the"
                    f" provided distance matrix has rows pertaining to input"
                    f" pattern and columns pertaining to the output points.",
                    stacklevel=2,
                )
                distances = distances.min(axis=0)
            elif k == n:
                distances = distances.min(axis=0)
            else:
                raise ValueError(
                    f"Distance matrix should have the same rows as the input"
                    f" coordinates with p columns, where n may be equal to p."
                    f" Recieved an {k},{p} distance matrix for {n} coordinates"
                )
        elif distances.ndim == 1:
            p = len(distances)
    else:
        # Do 1000 empties. Users can control this by computing their own
        # empty space distribution.
        n_empty_points = 1000

        randoms = poisson(hull=hull, size=(n_empty_points, 1))
        try:
            tree
        except NameError:
            tree = _build_best_tree(coordinates, metric)
        finally:
            distances, _ = tree.query(randoms, k=1)
            distances = distances.squeeze()

    counts, bins = numpy.histogram(distances, bins=support)
    fracs = numpy.cumsum(counts) / counts.sum()

    return bins, numpy.asarray([0, *fracs])


def g(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    edge_correction=None,
):
    """
    Ripley's G function

    The G function is computed from the cumulative density function of the nearest neighbor
    distances between points in the pattern.

    Parameters
    -----------
    coordinates : numpy.ndarray of shape (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, n) or (n,)
        distances from every point in the point to another point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.

    Returns
    -------
    a tuple containing the support values used to evalute the function
    and the values of the function at each distance value in the support.

    """

    coordinates, support, distances, metric, *_ = _prepare(
        coordinates, support, distances, metric, None, edge_correction
    )
    if distances is not None:
        if distances.ndim == 2:
            if distances.shape[0] == distances.shape[1] == coordinates.shape[0]:
                warnings.warn(
                    "The full distance matrix is not required for this function,"
                    " only the distance to the nearest neighbor within the pattern."
                    " Computing this and discarding the rest.",
                    stacklevel=2,
                )
                distances = distances.min(axis=1)
            else:
                k, p = distances.shape
                n = coordinates.shape[0]
                raise ValueError(
                    " Input distance matrix has an invalid shape: {k},{p}."
                    " Distances supplied can either be 2 dimensional"
                    " square matrices with the same number of rows"
                    " as `coordinates` ({n}) or 1 dimensional and contain"
                    " the shortest distance from each point in "
                    " `coordinates` to some other point in coordinates."
                )
        elif distances.ndim == 1:
            if distances.shape[0] != coordinates.shape[0]:
                raise ValueError(
                    f"Distances are not aligned with coordinates! Distance"
                    f" matrix must be (n_coordinates, n_coordinates), but recieved"
                    f" {distances.shape} instead of ({coordinates.shape[0]},)"
                )
        else:
            raise ValueError(
                "Distances supplied can either be 2 dimensional"
                " square matrices with the same number of rows"
                " as `coordinates` or 1 dimensional and contain"
                " the shortest distance from each point in "
                " `coordinates` to some other point in coordinates."
                " Input matrix was {distances.ndim} dimensioanl"
            )
    else:
        try:
            tree
        except NameError:
            tree = _build_best_tree(coordinates, metric)
        finally:
            distances, indices = _k_neighbors(tree, coordinates, k=1)

    counts, bins = numpy.histogram(distances.squeeze(), bins=support)
    fracs = numpy.cumsum(counts) / counts.sum()

    return bins, numpy.asarray([0, *fracs])


def j(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    truncate=True,
):
    """
    Ripely's J function

    The so-called "spatial hazard" function, this is a function relating the F and G functions.

    Parameters
    -----------
    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: tuple of numpy.ndarray
        precomputed distances to use to evaluate the j function.
        The first must be of shape (n,n) or (n,) and is used in the g function.
        the second must be of shape (n,p) or (p,) (with p possibly equal to n)
        used in the f function.
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern for the f function.
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    truncate: bool (default: True)
        whether or not to truncate the results when the F function reaches one. If the
        F function is one but the G function is less than one, this function will return
        numpy.nan values.

    Returns
    -------
    a tuple containing the support values used to evalute the function
    and the values of the function at each distance value in the support.
    """
    if distances is not None:
        g_distances, f_distances = distances
    else:
        g_distances = f_distances = None
    fsupport, fstats = f(
        coordinates,
        support=support,
        distances=f_distances,
        metric=metric,
        hull=hull,
        edge_correction=edge_correction,
    )

    gsupport, gstats = g(
        coordinates,
        support=support,
        distances=g_distances,
        metric=metric,
        edge_correction=edge_correction,
    )

    if isinstance(support, numpy.ndarray):
        if not numpy.allclose(gsupport, support):
            gfunction = interpolate.interp1d(gsupport, gstats, fill_value=1)
            gstats = gfunction(support)
            gsupport = support
    if not (numpy.allclose(gsupport, fsupport)):
        ffunction = interpolate.interp1d(fsupport, fstats, fill_value=1)
        fstats = ffunction(gsupport)
        fsupport = gsupport

    with numpy.errstate(invalid="ignore", divide="ignore"):
        hazard_ratio = (1 - gstats) / (1 - fstats)
    both_zero = (gstats == 1) & (fstats == 1)
    hazard_ratio[both_zero] = numpy.nan
    if truncate:
        result = _truncate(gsupport, hazard_ratio)
        if len(result[1]) != len(hazard_ratio):
            warnings.warn(
                f"requested {support} bins to evaluate the J function, but"
                f" it reaches infinity at d={result[0][-1]:.4f}, meaning only"
                f" {len(result[0])} bins will be used to characterize the J function.",
                stacklevel=2,
            )
        return result
    else:
        return gsupport, hazard_ratio


def k(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    edge_correction=None,
):
    """
    Ripley's K function

    This function counts the number of pairs of points that are closer than a given distance.
    As d increases, K approaches the number of point pairs.

    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.

    Returns
    -------
    a tuple containing the support values used to evalute the function
    and the values of the function at each distance value in the support.
    """
    coordinates, support, distances, metric, hull, edge_correction = _prepare(
        coordinates, support, distances, metric, None, edge_correction
    )
    n = coordinates.shape[0]
    upper_tri_n = n * (n - 1) * 0.5
    if distances is not None:
        if distances.ndim == 1:
            if distances.shape[0] != upper_tri_n:
                raise ValueError(
                    f"Shape of inputted distances is not square, nor is the upper triangular"
                    f" matrix matching the number of input points. The shape of the input matrix"
                    f" is {distances.shape}, but required shape is ({upper_tri_n},) or ({n},{n})"
                )
            upper_tri_distances = distances
        elif distances.shape[0] == distances.shape[1] == n:
            upper_tri_distances = distances[numpy.triu_indices_from(distances, k=1)]
        else:
            raise ValueError(
                f"Shape of inputted distances is not square, nor is the upper triangular"
                f" matrix matching the number of input points. The shape of the input matrix"
                f" is {distances.shape}, but required shape is ({upper_tri_n},) or ({n},{n})"
            )
    else:
        upper_tri_distances = spatial.distance.pdist(coordinates, metric=metric)
    n_pairs_less_than_d = (upper_tri_distances < support.reshape(-1, 1)).sum(axis=1)
    intensity = n / _area(hull)
    k_estimate = ((n_pairs_less_than_d * 2) / n) / intensity
    return support, k_estimate


def l(
    coordinates,
    support=None,
    permutations=9999,
    distances=None,
    metric="euclidean",
    edge_correction=None,
    linearized=False,
):
    """
    Ripley's L function

    This is a scaled and shifted version of the K function that accounts for the K function's
    increasing expected value as distances increase. This means that the L function, for a
    completely random pattern, should be close to zero at all distance values in the support.

    Parameters
    ----------
    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    linearized : bool
        whether or not to subtract l from its expected value (support) at each
        distance bin. This centers the l function on zero for all distances.
        Proposed by Besag (1977)

    Returns
    -------
    a tuple containing the support values used to evalute the function
    and the values of the function at each distance value in the support.
    """

    support, k_estimate = k(
        coordinates,
        support=support,
        distances=distances,
        metric=metric,
        edge_correction=edge_correction,
    )

    l = numpy.sqrt(k_estimate / numpy.pi)

    if linearized:
        return support, l - support
    return support, l


# ------------------------------------------------------------#
# Statistical Tests based on Ripley Functions                 #
# ------------------------------------------------------------#

FtestResult = namedtuple(
    "FtestResult", ("support", "statistic", "pvalue", "simulations")
)
GtestResult = namedtuple(
    "GtestResult", ("support", "statistic", "pvalue", "simulations")
)
JtestResult = namedtuple(
    "JtestResult", ("support", "statistic", "pvalue", "simulations")
)
KtestResult = namedtuple(
    "KtestResult", ("support", "statistic", "pvalue", "simulations")
)
LtestResult = namedtuple(
    "LtestResult", ("support", "statistic", "pvalue", "simulations")
)

_ripley_dispatch = {
    "F": (f, FtestResult),
    "G": (g, GtestResult),
    "J": (j, JtestResult),
    "K": (k, KtestResult),
    "L": (l, LtestResult),
}


def _ripley_test(
    calltype,
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    keep_simulations=False,
    n_simulations=9999,
    **kwargs,
):
    stat_function, result_container = _ripley_dispatch.get(calltype)
    core_kwargs = dict(
        support=support,
        metric=metric,
        edge_correction=edge_correction,
    )
    tree = _build_best_tree(coordinates, metric=metric)
    hull = _prepare_hull(coordinates, hull)
    if calltype in ("F", "J"):  # these require simulations
        core_kwargs["hull"] = hull
        # amortize to avoid doing this every time
        empty_space_points = poisson(coordinates, size=(1000, 1))
        if distances is None:
            empty_space_distances, _ = _k_neighbors(tree, empty_space_points, k=1)
            if calltype == "F":
                distances = empty_space_distances.squeeze()
            else:  # calltype == 'J':
                n_distances, _ = _k_neighbors(tree, coordinates, k=1)
                distances = (n_distances.squeeze(), empty_space_distances.squeeze())
        else:
            pass
    core_kwargs.update(**kwargs)

    observed_support, observed_statistic = stat_function(
        tree, distances=distances, **core_kwargs
    )
    core_kwargs["support"] = observed_support
    n_observations = coordinates.shape[0]

    if keep_simulations:
        simulations = numpy.empty((len(observed_support), n_simulations)).T
    pvalues = numpy.ones_like(observed_support)
    for i_replication in range(n_simulations):
        random_i = poisson(hull, size=n_observations)
        if calltype in ("F", "J"):
            random_tree = _build_best_tree(random_i, metric)
            empty_distances, _ = random_tree.query(empty_space_points, k=1)
            if calltype == "F":
                core_kwargs["distances"] = empty_distances.squeeze()
            else:  # calltype == 'J':
                n_distances, _ = _k_neighbors(random_tree, random_i, k=1)
                core_kwargs["distances"] = (
                    n_distances.squeeze(),
                    empty_distances.squeeze(),
                )
        rep_support, simulations_i = stat_function(random_i, **core_kwargs)
        pvalues += simulations_i >= observed_statistic
        if keep_simulations:
            simulations[i_replication] = simulations_i
    pvalues /= n_simulations + 1
    pvalues = numpy.minimum(pvalues, 1 - pvalues)
    return result_container(
        observed_support,
        observed_statistic,
        pvalues,
        simulations if keep_simulations else None,
    )


def f_test(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    keep_simulations=False,
    n_simulations=9999,
):
    """
    Ripley's F function

    The so-called "empty space" function, this is the cumulative density function of
    the distances from a random set of points to the known points in the pattern.

    When the estimated statistic is larger than simulated values at a given distance, then
    the pattern is considered "dispersed" or "regular"

    Parameters
    -----------
    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    keep_simulations: bool
        whether or not to keep the simulation envelopes. If so,
        will be returned as the result's simulations attribute
    n_simulations: int
        how many simulations to conduct, assuming that the reference pattern
        has complete spatial randomness.

    Returns
    -------
    a named tuple with properties
    - support, the exact distance values used to evalute the statistic
    - statistic, the values of the statistic at each distance
    - pvalue, the percent of simulations that were as extreme as the observed value
    - simulations, the distribution of simulated statistics (shaped (n_simulations, n_support_points))
        or None if keep_simulations=False (which is the default)
    """

    return _ripley_test(
        "F",
        coordinates,
        support=support,
        distances=distances,
        metric=metric,
        hull=hull,
        edge_correction=edge_correction,
        keep_simulations=keep_simulations,
        n_simulations=n_simulations,
    )


def g_test(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    keep_simulations=False,
    n_simulations=9999,
):
    """
    Ripley's G function

    The G function is computed from the cumulative density function of the nearest neighbor
    distances between points in the pattern.

    When the G function is below the simulated values, it suggests dispersion.

    Parameters
    ----------
    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    keep_simulations: bool
        whether or not to keep the simulation envelopes. If so,
        will be returned as the result's simulations attribute
    n_simulations: int
        how many simulations to conduct, assuming that the reference pattern
        has complete spatial randomness.

    Returns
    -------
    a named tuple with properties
    - support, the exact distance values used to evalute the statistic
    - statistic, the values of the statistic at each distance
    - pvalue, the percent of simulations that were as extreme as the observed value
    - simulations, the distribution of simulated statistics (shaped (n_simulations, n_support_points))
        or None if keep_simulations=False (which is the default)
    """
    return _ripley_test(
        "G",
        coordinates,
        support=support,
        distances=distances,
        metric=metric,
        hull=hull,
        edge_correction=edge_correction,
        keep_simulations=keep_simulations,
        n_simulations=n_simulations,
    )


def j_test(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    truncate=True,
    keep_simulations=False,
    n_simulations=9999,
):
    """
    Ripley's J function

    The so-called "spatial hazard" function, this is a function relating the F and G functions.

    When the J function is consistently below 1, then it indicates clustering.
    When consistently above 1, it suggests dispersion.

    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    keep_simulations: bool
        whether or not to keep the simulation envelopes. If so,
        will be returned as the result's simulations attribute
    n_simulations: int
        how many simulations to conduct, assuming that the reference pattern
        has complete spatial randomness.

    Returns
    -------
    a named tuple with properties
    - support, the exact distance values used to evalute the statistic
    - statistic, the values of the statistic at each distance
    - pvalue, the percent of simulations that were as extreme as the observed value
    - simulations, the distribution of simulated statistics (shaped (n_simulations, n_support_points))
        or None if keep_simulations=False (which is the default)
    """
    result = _ripley_test(
        "J",
        coordinates,
        support=support,
        distances=distances,
        metric=metric,
        hull=hull,
        edge_correction=edge_correction,
        keep_simulations=keep_simulations,
        n_simulations=n_simulations,
        truncate=False,
    )
    if truncate:
        result_trunc = _truncate(*result)
        result_trunc = JtestResult(*result_trunc)
        if len(result_trunc.statistic) != len(result.statistic):
            warnings.warn(
                f"requested {support} bins to evaluate the J function, but"
                f" it reaches infinity at d={result[0][-1]:.4f}, meaning only"
                f" {len(result[0])} bins will be used to characterize the J function.",
                stacklevel=2,
            )
            return result_trunc

    else:
        return result


def k_test(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    keep_simulations=False,
    n_simulations=9999,
):
    """
    Ripley's K function

    This function counts the number of pairs of points that are closer than a given distance.
    As d increases, K approaches the number of point pairs.

    When the K function is below simulated values, it suggests that the pattern is dispersed.

    Parameters
    ----------
    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    keep_simulations: bool
        whether or not to keep the simulation envelopes. If so,
        will be returned as the result's simulations attribute
    n_simulations: int
        how many simulations to conduct, assuming that the reference pattern
        has complete spatial randomness.

    Returns
    -------
    a named tuple with properties
    - support, the exact distance values used to evalute the statistic
    - statistic, the values of the statistic at each distance
    - pvalue, the percent of simulations that were as extreme as the observed value
    - simulations, the distribution of simulated statistics (shaped (n_simulations, n_support_points))
        or None if keep_simulations=False (which is the default)
    """
    return _ripley_test(
        "K",
        coordinates,
        support=support,
        distances=distances,
        metric=metric,
        hull=hull,
        edge_correction=edge_correction,
        keep_simulations=keep_simulations,
        n_simulations=n_simulations,
    )


def l_test(
    coordinates,
    support=None,
    distances=None,
    metric="euclidean",
    hull=None,
    edge_correction=None,
    linearized=False,
    keep_simulations=False,
    n_simulations=9999,
):
    """
    Ripley's L function

    This is a scaled and shifted version of the K function that accounts for the K function's
    increasing expected value as distances increase. This means that the L function, for a
    completely random pattern, should be close to zero at all distance values in the support.

    When the L function is negative, this suggests dispersion.

    Parameters
    ----------
    coordinates : numpy.ndarray, (n,2)
        input coordinates to function
    support : tuple of length 1, 2, or 3, int, or numpy.ndarray
        tuple, encoding (stop,), (start, stop), or (start, stop, num)
        int, encoding number of equally-spaced intervals
        numpy.ndarray, used directly within numpy.histogram
    distances: numpy.ndarray, (n, p) or (p,)
        distances from every point in a random point set of size p
        to some point in `coordinates`
    metric: str or callable
        distance metric to use when building search tree
    hull: bounding box, scipy.spatial.ConvexHull, shapely.geometry.Polygon
        the hull used to construct a random sample pattern, if distances is None
    edge_correction: bool or str
        whether or not to conduct edge correction. Not yet implemented.
    keep_simulations: bool
        whether or not to keep the simulation envelopes. If so,
        will be returned as the result's simulations attribute
    n_simulations: int
        how many simulations to conduct, assuming that the reference pattern
        has complete spatial randomness.

    Returns
    -------
    a named tuple with properties
    - support, the exact distance values used to evalute the statistic
    - statistic, the values of the statistic at each distance
    - pvalue, the percent of simulations that were as extreme as the observed value
    - simulations, the distribution of simulated statistics (shaped (n_simulations, n_support_points))
        or None if keep_simulations=False (which is the default)
    """
    return _ripley_test(
        "L",
        coordinates,
        support=support,
        distances=distances,
        metric=metric,
        hull=hull,
        edge_correction=edge_correction,
        linearized=linearized,
        keep_simulations=keep_simulations,
        n_simulations=n_simulations,
    )


def _truncate(support, realizations, *rest):
    is_invalid = numpy.isinf(realizations) | numpy.isnan(realizations)
    first_inv = is_invalid.argmax()
    if not is_invalid.any():
        return support, realizations, *rest
    elif first_inv < len(realizations):
        return (
            support[:first_inv],
            realizations[:first_inv],
            *[r[:first_inv] if r is not None else None for r in rest],
        )



"""
Centrographic measures for point patterns

TODO

- testing
- documentation

"""

__author__ = "Serge Rey sjsrey@gmail.com"

__all__ = [
    "mbr",
    "hull",
    "mean_center",
    "weighted_mean_center",
    "manhattan_median",
    "std_distance",
    "euclidean_median",
    "ellipse",
    "minimum_rotated_rectangle",
    "minimum_bounding_rectangle",
    "skyum",
    "dtot",
    "_circle",
]


import sys
import numpy as np
import warnings
import copy

from math import pi as PI
from scipy.spatial import ConvexHull
from libpysal.cg import get_angle_between, Ray, is_clockwise
from scipy.spatial import distance as dist
from scipy.optimize import minimize

not_clockwise = lambda x: not is_clockwise(x)

MAXD = sys.float_info.max
MIND = sys.float_info.min


def minimum_bounding_rectangle(points):
    """
    Find minimum bounding rectangle of a point array.

    Parameters
    ----------
    points : arraylike
             (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    min_x  : float
             leftmost value of the vertices of minimum bounding rectangle.
    min_y  : float
             downmost value of the vertices of minimum bounding rectangle.
    max_x  : float
             rightmost value of the vertices of minimum bounding rectangle.
    max_y  : float
             upmost value of the vertices of minimum bounding rectangle.

    """
    points = np.asarray(points)
    min_x = min_y = MAXD
    max_x = max_y = MIND
    x, y = zip(*points)
    min_x = min(x)
    min_y = min(y)
    max_x = max(x)
    max_y = max(y)
    return min_x, min_y, max_x, max_y


def minimum_rotated_rectangle(points, return_angle=False):
    """
    Compute the minimum rotated rectangle for an input point set.

    This is the smallest enclosing rectangle (possibly rotated)
    for the input point set. It is computed using OpenCV2, so
    if that is not available, then this function will fail.

    Parameters
    ----------
    points : numpy.ndarray
        A numpy array of shape (n_observations, 2) containing the point
        locations to compute the rotated rectangle
    return_angle : bool
        whether to return the angle (in degrees) of the angle between
        the horizontal axis of the rectanle and the first side (i.e. length).
        Computed directly from cv2.minAreaRect.

    Returns
    -------
    an numpy.ndarray of shape (4, 2) containing the coordinates
    of the minimum rotated rectangle. If return_angle is True,
    also return the angle (in degrees) of the rotated rectangle.

    """
    points = np.asarray(points)
    try:
        from cv2 import minAreaRect, boxPoints
    except (ImportError, ModuleNotFoundError):
        raise ModuleNotFoundError("OpenCV2 is required to use this function. Please install it with `pip install opencv-contrib-python`")
    ((x, y), (w, h), angle) = rot_rect = minAreaRect(points.astype(np.float32))
    out_points = boxPoints(rot_rect)
    if return_angle:
        return (out_points, angle)
    return out_points


def mbr(points):
    warnings.warn(
        "This function will be deprecated in the next release of pointpats.",
        FutureWarning,
        stacklevel=2,
    )
    return minimum_bounding_rectangle(points)


mbr.__doc__ = minimum_bounding_rectangle.__doc__


def hull(points):
    """
    Find convex hull of a point array.

    Parameters
    ----------
    points: arraylike
            (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _     : array
            (h,2), points defining the hull in counterclockwise order.
    """

    points = np.asarray(points)
    h = ConvexHull(points)
    return points[h.vertices]


def mean_center(points):
    """
    Find mean center of a point array.

    Parameters
    ----------
    points: arraylike
            (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _     : array
            (2,), (x,y) coordinates of the mean center.
    """

    points = np.asarray(points)
    return points.mean(axis=0)


def weighted_mean_center(points, weights):
    """
    Find weighted mean center of a marked point pattern.

    Parameters
    ----------
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.
    weights : arraylike
              a series of attribute values of length n.

    Returns
    -------
    _      : array
             (2,), (x,y) coordinates of the weighted mean center.
    """

    points, weights = np.asarray(points), np.asarray(weights)
    w = weights * 1.0 / weights.sum()
    w.shape = (1, len(points))
    return np.dot(w, points)[0]


def manhattan_median(points):
    """
    Find manhattan median of a point array.

    Parameters
    ----------
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _      : array
             (2,), (x,y) coordinates of the manhattan median.
    """

    points = np.asarray(points)
    if not len(points) % 2:
        s = "Manhattan Median is not unique for even point patterns."
        warnings.warn(s)
    return np.median(points, axis=0)


def std_distance(points):
    """
    Calculate standard distance of a point array.

    Parameters
    ----------
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _      : float
             standard distance.
    """

    points = np.asarray(points)
    n, p = points.shape
    m = points.mean(axis=0)
    return np.sqrt(((points * points).sum(axis=0) / n - m * m).sum())


def ellipse(points):
    """
    Calculate parameters of standard deviational ellipse for a point pattern.

    Parameters
    ----------
    points : arraylike
             (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _      : float
             semi-major axis.
    _      : float
             semi-minor axis.
    theta  : float
             clockwise rotation angle of the ellipse.

    Notes
    -----
    Implements approach from:

    https://www.icpsr.umich.edu/CrimeStat/files/CrimeStatChapter.4.pdf
    """

    points = np.asarray(points)
    n, k = points.shape
    x = points[:, 0]
    y = points[:, 1]
    xd = x - x.mean()
    yd = y - y.mean()
    xss = (xd * xd).sum()
    yss = (yd * yd).sum()
    cv = (xd * yd).sum()
    num = (xss - yss) + np.sqrt((xss - yss) ** 2 + 4 * (cv) ** 2)
    den = 2 * cv
    theta = np.arctan(num / den)
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)
    n_2 = n - 2
    sd_x = (2 * (xd * cos_theta - yd * sin_theta) ** 2).sum() / n_2
    sd_y = (2 * (xd * sin_theta - yd * cos_theta) ** 2).sum() / n_2
    return np.sqrt(sd_x), np.sqrt(sd_y), theta


def dtot(coord, points):
    """
    Sum of Euclidean distances between event points and a selected point.

    Parameters
    ----------
    coord   : arraylike
              (x,y) coordinates of a point.
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    d       : float
              sum of Euclidean distances.

    """
    points = np.asarray(points)
    xd = points[:, 0] - coord[0]
    yd = points[:, 1] - coord[1]
    d = np.sqrt(xd * xd + yd * yd).sum()
    return d


def euclidean_median(points):
    """
    Calculate the Euclidean median for a point pattern.

    Parameters
    ----------
    points: arraylike
            (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _     : array
            (2,), (x,y) coordinates of the Euclidean median.

    """
    points = np.asarray(points)
    start = mean_center(points)
    res = minimize(dtot, start, args=(points,))
    return res["x"]


def minimum_bounding_circle(points):
    """
    Implements Skyum (1990)'s algorithm for the minimum bounding circle in R^2.

    Store points clockwise.
    Find p in S that maximizes angle(prec(p), p, succ(p) THEN radius(prec(
    p), p, succ(p)). This is also called the lexicographic maximum, and is the last
    entry of a list of (radius, angle) in lexicographical order.

    * If angle(prec(p), p, succ(p)) <= 90 degrees, then finish.

    * If not, remove p from set.

    Parameters
    ----------
    points  :   numpy.ndarray
        a numpy array of shape (n_observations, 2) to compute
        the minimum bounding circle

    Returns
    -------
    (x,y),center for the minimum bounding circle.
    """
    points = hull(points)
    if not_clockwise(points):
        points = points[::-1]
        if not_clockwise(points):
            raise Exception("Points are neither clockwise nor counterclockwise")
    POINTS = copy.deepcopy(points)
    removed = []
    i = 0
    if HAS_NUMBA:
        circ = _skyum_numba(POINTS)[0]
    else:
        circ = _skyum_lists(POINTS)[0]
    return (circ[1], circ[2]), circ[0]


def skyum(points):
    warnings.warn(
        "This function will be deprecated in the next release of pointpats.",
        FutureWarning,
        stacklevel=2,
    )
    return minimum_bounding_circle(points)


skyum.__doc__ = (
    "WARNING: This function is deprecated in favor of minimum_bounding_circle\n"
    + minimum_bounding_circle.__doc__
)


def _skyum_lists(points):
    points = points.tolist()
    removed = []
    i = 0
    while True:
        angles = [_angle(_prec(p, points), p, _succ(p, points),) for p in points]
        circles = [_circle(_prec(p, points), p, _succ(p, points),) for p in points]
        radii = [c[0] for c in circles]
        lexord = np.lexsort((radii, angles))  # confusing as hell defaults...
        lexmax = lexord[-1]
        candidate = (
            _prec(points[lexmax], points),
            points[lexmax],
            _succ(points[lexmax], points),
        )
        if angles[lexmax] <= (np.pi / 2.0):
            # print("Constrained by points: {}".format(candidate))
            return _circle(*candidate), points, removed, candidate
        else:
            try:
                removed.append((points.pop(lexmax), i))
            except IndexError:
                raise Exception("Construction of Minimum Bounding Circle failed!")
        i += 1


try:
    from numba import njit, boolean

    HAS_NUMBA = True

    @njit(fastmath=True)
    def _skyum_numba(points):
        i = 0
        complete = False
        while not complete:
            complete, points, candidate, circle = _skyum_iteration(points)
            if complete:
                return circle, points, None, candidate

    @njit(fastmath=True)
    def _skyum_iteration(points):
        points = points.reshape(-1, 2)
        n = points.shape[0]
        angles = np.empty((n,))
        circles = np.empty((n, 3))
        for i in range(n):
            p = points[(i - 1) % n]
            q = points[i % n]
            r = points[(i + 1) % n]
            angles[i] = _angle(p, q, r)
            circles[i] = _circle(p, q, r)
        radii = circles[:, 0]
        # workaround for no lexsort in numba
        angle_argmax = angles.argmax()
        angle_max = angles[angle_argmax]
        # the maximum radius for the largest angle
        lexmax = (radii * (angles == angle_max)).argmax()

        candidate = (lexmax - 1) % n, lexmax, (lexmax + 1) % n
        if angles[lexmax] <= (np.pi / 2.0):
            return True, points, lexmax, circles[lexmax]
        else:
            mask = np.ones((n,), dtype=boolean)
            mask[lexmax] = False
            new_points = points[mask, :]
            return False, new_points, lexmax, circles[lexmax]


except ModuleNotFoundError:

    def njit(func, **kwargs):
        return func


@njit
def _angle(p, q, r):
    pq = p - q
    rq = r - q
    magnitudes = np.linalg.norm(pq) * np.linalg.norm(rq)
    return np.abs(np.arccos(np.dot(pq, rq) / magnitudes))


def _prec(p, l):
    """
    retrieve the predecessor of p in list l
    """
    pos = l.index(p)
    if pos - 1 < 0:
        return l[-1]
    else:
        return l[pos - 1]


def _succ(p, l):
    """
    retrieve the successor of p in list l
    """
    pos = l.index(p)
    if pos + 1 >= len(l):
        return l[0]
    else:
        return l[pos + 1]


@njit
def _euclidean_distance(px, py, qx, qy):
    return np.sqrt((px - qx) ** 2 + (py - qy) ** 2)


@njit
def _circle(p, q, r, dmetric=_euclidean_distance):
    """
    Returns (radius, (center_x, center_y)) of the circumscribed circle by the
    triangle pqr.

    note, this does not assume that p!=q!=r
    """
    px, py = p
    qx, qy = q
    rx, ry = r
    angle = np.abs(_angle(p, q, r))
    if np.abs(angle - np.pi) < 1e-5:  # angle is pi
        radius = _euclidean_distance(px, py, rx, ry) / 2.0
        center_x = (px + rx) / 2.0
        center_y = (py + ry) / 2.0
    elif np.abs(angle) < 1e-5:  # angle is zero
        radius = _euclidean_distance(px, py, qx, qy) / 2.0
        center_x = (px + qx) / 2.0
        center_y = (py + qy) / 2.0
    else:
        D = 2 * (px * (qy - ry) + qx * (ry - py) + rx * (py - qy))
        center_x = (
            (px ** 2 + py ** 2) * (qy - ry)
            + (qx ** 2 + qy ** 2) * (ry - py)
            + (rx ** 2 + ry ** 2) * (py - qy)
        ) / float(D)
        center_y = (
            (px ** 2 + py ** 2) * (rx - qx)
            + (qx ** 2 + qy ** 2) * (px - rx)
            + (rx ** 2 + ry ** 2) * (qx - px)
        ) / float(D)
        radius = _euclidean_distance(center_x, center_y, px, py)
    return radius, center_x, center_y



import numpy
from .geometry import (
    spatial,
    area as _area,
    centroid as _centroid,
    contains as _contains,
    bbox as _bbox,
    prepare_hull as _prepare_hull,
    HULL_TYPES,
)

# ------------------------------------------------------------ #
# Utilities                                                    #
# ------------------------------------------------------------ #


def parse_size_and_intensity(hull, intensity=None, size=None):
    """
    Given a hull, an intensity, and a size int/tuple, correctly
    compute the resulting missing quantities. Defaults to 100 points in one
    replication, meaning the intensity will be computed on the fly
    if nothing is provided.

    Parameters
    ----------
    hull : A geometry-like object
        This encodes the "space" in which to simulate the normal pattern. All points will
        lie within this hull. Supported values are:
        - a bounding box encoded in a numpy array as numpy.array([xmin, ymin, xmax, ymax])
        - an (N,2) array of points for which the bounding box will be computed & used
        - a shapely polygon/multipolygon
        - a scipy convexh hull
    intensity : float
        the number of observations per unit area in the hull to use. If provided,
        then the number of observations is determined using the intensity * area(hull) and
        the size is assumed to represent n_replications (if provided).
    size : tuple or int
        a tuple of (n_observations, n_replications), where the first number is the number
        of points to simulate in each replication and the second number is the number of
        total replications. So, (10, 4) indicates 10 points, 4 times.
        If an integer is provided and intensity is None, n_replications is assumed to be 1.
        If size is an integer and intensity is also provided, then size indicates n_replications,
        and the number of observations is computed on the fly using intensity and area.
    """
    if size is None:
        if intensity is not None:
            # if intensity is provided, assume
            # n_observations
            n_observations = int(_area(hull) * intensity)
        else:
            # default to 100 points
            n_observations = 100
            intensity = n_observations / _area(hull)
        n_simulations = 1
        size = (n_observations, n_simulations)
    elif isinstance(size, tuple):
        if len(size) == 2 and intensity is None:
            n_observations, n_simulations = size
            intensity = n_observations / _area(hull)
        elif len(size) == 2 and intensity is not None:
            raise ValueError(
                "Either intensity or size as (n observations, n simulations)"
                " can be provided. Providing both creates statistical conflicts."
                " between the requested intensity and implied intensity by"
                " the number of observations and the area of the hull. If"
                " you want to specify the intensity, use the intensity argument"
                " and set size equal to the number of simulations."
            )
        else:
            raise ValueError(
                f"Intensity and size not understood. Provide size as a tuple"
                f" containing (number of observations, number of simulations)"
                f" with no specified intensity, or an intensity and size equal"
                f" to the number of simulations."
                f" Recieved: `intensity={intensity}, size={size}`"
            )
    elif isinstance(size, int):
        # assume int size with specified intensity means n_simulations at x intensity
        if intensity is not None:
            n_observations = int(intensity * _area(hull))
            n_simulations = size
        else:  # assume we have one replication at the specified number of points
            n_simulations = 1
            n_observations = size
            intensity = n_observations / _area(hull)
    else:
        raise ValueError(
            f"Intensity and size not understood. Provide size as a tuple"
            f" containing (number of observations, number of simulations)"
            f" with no specified intensity, or an intensity and size equal"
            f" to the number of simulations."
            f" Recieved: `intensity={intensity}, size={size}`"
        )
    return (n_observations, n_simulations, intensity)


# ------------------------------------------------------------ #
# Distributions                                                #
# ------------------------------------------------------------ #


def poisson(hull, intensity=None, size=None):
    """
    Simulate a poisson random point process with a specified intensity.

    Parameters
    ----------
    hull : A geometry-like object
        This encodes the "space" in which to simulate the normal pattern. All points will
        lie within this hull. Supported values are:
        - a bounding box encoded in a numpy array as numpy.array([xmin, ymin, xmax, ymax])
        - an (N,2) array of points for which the bounding box will be computed & used
        - a shapely polygon/multipolygon
        - a scipy convexh hull
    intensity : float
        the number of observations per unit area in the hull to use. If provided, then
        size must be an integer describing the number of replications to use.
    size : tuple or int
        a tuple of (n_observations, n_replications), where the first number is the number
        of points to simulate in each replication and the second number is the number of
        total replications. So, (10, 4) indicates 10 points, 4 times.
        If an integer is provided and intensity is None, n_replications is assumed to be 1.
        If size is an integer and intensity is also provided, then size indicates n_replications,
        and the number of observations is computed from the intensity.

    Returns
    --------
        :   numpy.ndarray
        either an (n_replications, n_observations, 2) or (n_observations,2) array containing
        the simulated realizations.
    """
    if isinstance(hull, numpy.ndarray):
        if hull.shape == (4,):
            hull = hull
        else:
            hull = _prepare_hull(hull)
    n_observations, n_simulations, intensity = parse_size_and_intensity(
        hull, intensity=intensity, size=size
    )

    result = numpy.empty((n_simulations, n_observations, 2))

    bbox = _bbox(hull)

    for i_replication in range(n_simulations):
        generating = True
        i_observation = 0
        while i_observation < n_observations:
            x, y = (
                numpy.random.uniform(bbox[0], bbox[2]),
                numpy.random.uniform(bbox[1], bbox[3]),
            )
            if _contains(hull, x, y):
                result[i_replication, i_observation] = (x, y)
                i_observation += 1
    return result.squeeze()


def normal(hull, center=None, cov=None, size=None):
    """
    Simulate a multivariate random normal point cluster

    Parameters
    ----------
    hull : A geometry-like object
        This encodes the "space" in which to simulate the normal pattern. All points will
        lie within this hull. Supported values are:
        - a bounding box encoded in a numpy array as numpy.array([xmin, ymin, xmax, ymax])
        - an (N,2) array of points for which the bounding box will be computed & used
        - a shapely polygon/multipolygon
        - a scipy convexh hull
    center : iterable of shape (2, )
        A point where the simulations will be centered.
    cov : float or a numpy array of shape (2,2)
        either the standard deviation of an independent and identically distributed
        normal distribution, or a 2 by 2 covariance matrix expressing the covariance
        of the x and y for the distribution. Default is half of the width or height
        of the hull's bounding box, whichever is larger.
    size : tuple or int
        a tuple of (n_observations, n_replications), where the first number is the number
        of points to simulate in each replication and the second number is the number of
        total replications. So, (10, 4) indicates 10 points, 4 times.
        If an integer is provided, n_replications is assumed to be 1.

    Returns
    --------
        :   numpy.ndarray
        either an (n_replications, n_observations, 2) or (n_observations,2) array containing
        the simulated realizations.
    """
    if isinstance(hull, numpy.ndarray):
        if hull.shape == (4,):
            hull = hull
        else:
            hull = _prepare_hull(hull)
    if center is None:
        center = _centroid(hull)
    n_observations, n_simulations, intensity = parse_size_and_intensity(
        hull, intensity=None, size=size
    )
    if cov is None:
        bbox = _bbox(hull)
        width = bbox[2] - bbox[0]
        height = bbox[3] - bbox[1]
        cov = numpy.maximum(width / 2, height / 2) ** 2

    if isinstance(cov, (int, float)):
        sd = cov
        cov = numpy.eye(2) * sd
    elif isnstance(cov, numpy.ndarray):
        if cov.ndim == 2:
            assert cov.shape == (2, 2), "Bivariate covariance matrices must be 2 by 2"
        elif cov.ndim == 3:
            assert cov.shape[1:] == (
                2,
                2,
            ), "3-dimensional covariance matrices must have shape (n_simulations, 2,2)"
            assert (
                cov.shape[0] == n_simulations
            ), "3-dimensional covariance matrices must have shape (n_simulations, 2,2)"
    else:
        raise ValueError(
            "`cov` argument must be a float (signifying a standard deviation)"
            " or a 2 by 2 array expressing the covariance matrix of the "
            " bivariate normal distribution."
        )

    result = numpy.empty((n_simulations, n_observations, 2))

    bbox = _bbox(hull)

    for i_replication in range(n_simulations):
        generating = True
        i_observation = 0
        replication_cov = cov[i] if cov.ndim == 3 else cov
        replication_sd = numpy.diagonal(replication_cov) ** 0.5
        replication_cor = (1 / replication_sd) * replication_cov * (1 / replication_sd)

        while i_observation < n_observations:
            candidate = numpy.random.multivariate_normal((0, 0), replication_cor)
            x, y = center + candidate * replication_sd
            if _contains(hull, x, y):
                result[i_replication, i_observation] = (x, y)
                i_observation += 1
    return result.squeeze()


def cluster_poisson(
    hull, intensity=None, size=None, n_seeds=2, cluster_radius=None,
):
    """
    Simulate a cluster poisson random point process with a specified intensity & number of seeds.
    A cluster poisson process is a poisson process where the center of each "cluster" is
    itself distributed according to a spatial poisson process.

    Parameters
    ----------
    hull : A geometry-like object
        This encodes the "space" in which to simulate the normal pattern. All points will
        lie within this hull. Supported values are:
        - a bounding box encoded in a numpy array as numpy.array([xmin, ymin, xmax, ymax])
        - an (N,2) array of points for which the bounding box will be computed & used
        - a shapely polygon/multipolygon
        - a scipy convexh hull
    intensity : float
        the number of observations per unit area in the hull to use. If provided, then
        size must be an integer describing the number of replications to use.
    size : tuple or int
        a tuple of (n_observations, n_replications), where the first number is the number
        of points to simulate in each replication and the second number is the number of
        total replications. So, (10, 4) indicates 10 points, 4 times.
        If an integer is provided and intensity is None, n_replications is assumed to be 1.
        If size is an integer and intensity is also provided, then size indicates n_replications,
        and the number of observations is computed from the intensity.
    n_seeds : int
        the number of sub-clusters to use.
    cluster_radius : float or iterable
        the radius of each cluster. If a float, the same radius is used for all clusters.
        If an array, then there must be the same number of radii as clusters.
        If None, 50% of the minimum inter-point distance is used, which may fluctuate across
        replications.

    Returns
    --------
        :   numpy.ndarray
        either an (n_replications, n_observations, 2) or (n_observations,2) array containing
        the simulated realizations.
    """
    if isinstance(hull, numpy.ndarray):
        if hull.shape == (4,):
            hull = hull
        else:
            hull = _prepare_hull(hull)

    if isinstance(cluster_radius, numpy.ndarray):
        cluster_radii = cluster_radius.flatten()
        assert len(cluster_radii) == n_seeds, (
            f"number of radii provided ({len(cluster_radii)})"
            f"does not match number of clusters requested"
            f" ({n_seeds})."
        )
    elif isinstance(cluster_radius, (int, float)):
        cluster_radii = [cluster_radius] * n_seeds

    n_observations, n_simulations, intensity = parse_size_and_intensity(
        hull, intensity=intensity, size=size
    )

    result = numpy.empty((n_simulations, n_observations, 2))
    hull_area = _area(hull)
    for i_replication in range(n_simulations):
        seeds = poisson(hull, size=(n_seeds, n_simulations))
        if cluster_radius is None:
            # default cluster radius is one half the minimum distance between seeds
            cluster_radii = [spatial.distance.pdist(seeds).min() * 0.5] * n_seeds
        clusters = numpy.array_split(result[i_replication], n_seeds)
        for i_cluster, radius in enumerate(cluster_radii):
            seed = seeds[i_cluster]
            cluster_points = clusters[i_cluster]
            n_in_cluster = len(cluster_points)
            if n_in_cluster == 1:
                clusters[i_cluster] = seed
                continue
            if n_in_cluster < 1:
                raise Exception(
                    "There are too many clusters requested for the "
                    " inputted number of samples. Reduce `n_seeds` or"
                    " increase the number of sampled points."
                )
            candidates = _uniform_circle(
                n_in_cluster - 1, radius=radius, center=seed, hull=hull
            )
            clusters[i_cluster] = numpy.row_stack((seed, candidates))
        result[i_replication] = numpy.row_stack(clusters)
    return result.squeeze()


def cluster_normal(hull, cov=None, size=None, n_seeds=2):
    """
    Simulate a cluster poisson random point process with a specified intensity & number of seeds.
    A cluster poisson process is a poisson process where the center of each "cluster" is
    itself distributed according to a spatial poisson process.

    Parameters
    ----------
    hull : A geometry-like object
        This encodes the "space" in which to simulate the normal pattern. All points will
        lie within this hull. Supported values are:
        - a bounding box encoded in a numpy array as numpy.array([xmin, ymin, xmax, ymax])
        - an (N,2) array of points for which the bounding box will be computed & used
        - a shapely polygon/multipolygon
        - a scipy convexh hull
    cov : float, int, or numpy.ndarray of shape (2,2)
        The covariance structure for clusters. By default, this is the squared
        average distance between cluster seeds.
    size : tuple or int
        a tuple of (n_observations, n_replications), where the first number is the number
        of points to simulate in each replication and the second number is the number of
        total replications. So, (10, 4) indicates 10 points, 4 times.
        If an integer is provided and intensity is None, n_replications is assumed to be 1.
        If size is an integer and intensity is also provided, then size indicates n_replications,
        and the number of observations is computed from the intensity.
    n_seeds : int
        the number of sub-clusters to use.

    Returns
    --------
        :   numpy.ndarray
        either an (n_replications, n_observations, 2) or (n_observations,2) array containing
        the simulated realizations.
    """
    if isinstance(hull, numpy.ndarray):
        if hull.shape == (4,):
            hull = hull
        else:
            hull = _prepare_hull(hull)
    n_observations, n_simulations, intensity = parse_size_and_intensity(
        hull, intensity=None, size=size
    )
    result = numpy.empty((n_simulations, n_observations, 2))
    for i_replication in range(n_simulations):
        seeds = poisson(hull, size=(n_seeds, n_simulations))
        if cov is None:
            cov = spatial.distance.pdist(seeds).mean() ** 2
        clusters = numpy.array_split(result[i_replication], n_seeds)
        for i_cluster, seed in enumerate(seeds):
            cluster_points = clusters[i_cluster]
            n_in_cluster = len(cluster_points)
            if n_in_cluster == 1:
                clusters[i_cluster] = seed
                continue
            if n_in_cluster < 1:
                raise Exception(
                    "There are too many clusters requested for the "
                    " inputted number of samples. Reduce `n_seeds` or"
                    " increase the number of sampled points."
                )
            candidates = normal(hull, center=seed, cov=cov, size=n_in_cluster - 1)
            clusters[i_cluster] = numpy.row_stack((seed, candidates))
        result[i_replication] = numpy.row_stack(clusters)
    return result.squeeze()


def _uniform_circle(n, radius=1.0, center=(0.0, 0.0), burn=2, verbose=False, hull=None):
    """
    Generate n points within a circle of given radius.

    Parameters
    ----------
    n : int
        Number of points.
    radius : float
        Radius of the circle.
    center : tuple
        Coordinates of the center.
    burn : int
        number of coordinates to simulate at a time. This is the "chunk"
        size sent to numpy.random.uniform each iteration of the rejection sampler

    Returns
    -------
      : array
        (n, 2), coordinates of generated points

    """

    good = numpy.zeros((n, 2), float)
    c = 0
    center_x, center_y = center
    r = radius
    r2 = r * r
    it = 0
    while c < n:
        x = numpy.random.uniform(-r, r, (burn * n, 1))
        y = numpy.random.uniform(-r, r, (burn * n, 1))
        if hull is None:
            in_hull = True
        else:
            in_hull = numpy.asarray(
                [
                    _contains(hull, xi + center_x, yi + center_y)
                    for xi, yi in numpy.column_stack((x, y))
                ]
            ).reshape(-1, 1)
        ids, *_ = numpy.where(((x * x + y * y) <= r2) & in_hull)
        candidates = numpy.hstack((x, y))[ids]
        nc = candidates.shape[0]
        need = n - c
        if nc > need:  # more than we need
            good[c:] = candidates[:need]
        else:  # use them all and keep going
            good[c : c + nc] = candidates
        c += nc
        it += 1
    if verbose:
        print("Iterations: {}".format(it))
    return good + numpy.asarray(center)



import numpy
from scipy import spatial
from functools import singledispatch
from libpysal.cg import alpha_shape_auto
from libpysal.cg.kdtree import Arc_KDTree
import warnings

__all__ = ["area", "bbox", "contains", "k_neighbors", "build_best_tree", "prepare_hull"]

# ------------------------------------------------------------#
# Utilities and dispatching                                   #
# ------------------------------------------------------------#

TREE_TYPES = (spatial.KDTree, Arc_KDTree)
try:
    from sklearn.neighbors import KDTree, BallTree

    TREE_TYPES = (*TREE_TYPES, KDTree, BallTree)
except ModuleNotFoundError:
    pass

HULL_TYPES = (
    numpy.ndarray,
    spatial.ConvexHull,
)

## Define default dispatches and special dispatches without GEOS


### AREA
@singledispatch
def area(shape):
    """
    If a shape has an area attribute, return it.
    Works for:
        shapely.geometry.Polygon
    """
    try:
        return shape.area
    except AttributeError:
        return area(numpy.asarray(shape))


@area.register
def _(shape: spatial.ConvexHull):
    """
    If a shape is a convex hull from scipy,
    assure it's 2-dimensional and then use its volume.
    """
    assert shape.points.shape[1] == 2
    return shape.volume


@area.register
def _(shape: numpy.ndarray):
    """
    If a shape describes a bounding box, compute length times width
    """
    assert len(shape) == 4, "shape is not a bounding box!"
    width, height = shape[2] - shape[0], shape[3] - shape[1]
    return numpy.abs(width * height)


### bounding box
@singledispatch
def bbox(shape):
    """
    If a shape can be cast to an array, use that.
    Works for:
        lists of tuples
        scikit memory arrays
    """
    return bbox(numpy.asarray(shape))


@bbox.register
def _(shape: numpy.ndarray):
    """
    If a shape is an array of points, compute the minima/maxima
    or let it pass through if it's 1 dimensional & length 4
    """
    if (shape.ndim == 1) & (len(shape) == 4):
        return shape
    return numpy.array([*shape.min(axis=0), *shape.max(axis=0)])


@bbox.register
def _(shape: spatial.ConvexHull):
    """
    For scipy.spatial.ConvexHulls, compute the bounding box from
    their boundary points.
    """
    return bbox(shape.points[shape.vertices])


### contains


@singledispatch
def contains(shape, x, y):
    """
    Try to use the shape's contains method directly on XY.
    Does not currently work on anything.
    """
    raise NotImplementedError()
    return shape.contains((x, y))


@contains.register
def _(shape: numpy.ndarray, x: float, y: float):
    """
    If provided an ndarray, assume it's a bbox
    and return whether the point falls inside
    """
    xmin, xmax = shape[0], shape[2]
    ymin, ymax = shape[1], shape[3]
    in_x = (xmin <= x) and (x <= xmax)
    in_y = (ymin <= y) and (y <= ymax)
    return in_x & in_y


@contains.register
def _(shape: spatial.Delaunay, x: float, y: float):
    """
    For points and a delaunay triangulation, use the find_simplex
    method to identify whether a point is inside the triangulation.

    If the returned simplex index is -1, then the point is not
    within a simplex of the triangulation.
    """
    return shape.find_simplex((x, y)) >= 0


@contains.register
def _(shape: spatial.ConvexHull, x: float, y: float):
    """
    For convex hulls, convert their exterior first into a Delaunay triangulation
    and then use the delaunay dispatcher.
    """
    exterior = shape.points[shape.vertices]
    delaunay = spatial.Delaunay(exterior)
    return contains(delaunay, x, y)


### centroid
@singledispatch
def centroid(shape):
    """
    Assume the input is a shape with a centroid method:
    """
    return shape.centroid


@centroid.register
def _(shape: numpy.ndarray):
    """
    Handle point arrays or bounding boxes
    """
    from .centrography import mean_center

    if shape.ndim == 2:
        return mean_center(shape).squeeze()
    elif shape.ndim == 1:
        assert shape.shape == (4,)
        xmin, ymin, xmax, ymax = shape
        return numpy.column_stack(
            (numpy.mean((xmin, xmax)), numpy.mean((ymin, ymax)))
        ).squeeze()
    else:
        raise TypeError(
            f"Centroids are only implemented in 2 dimensions,"
            f" but input has {shape.ndim} dimensinos"
        )


@centroid.register
def _(shape: spatial.ConvexHull):
    """
    Treat convex hulls as arrays of points
    """
    return centroid(shape.points[shape.vertices])


try:
    from shapely.geometry.base import BaseGeometry as _BaseGeometry
    from shapely.geometry import (
        Polygon as _ShapelyPolygon,
        MultiPolygon as _ShapelyMultiPolygon,
    )
    from shapely.geometry import Point as _ShapelyPoint

    HULL_TYPES = (*HULL_TYPES, _ShapelyPolygon, _ShapelyMultiPolygon)
    HAS_SHAPELY = True

    @contains.register
    def _(shape: _BaseGeometry, x: float, y: float):
        """
        If we know we're working with a shapely polygon,
        then use the contains method & cast input coords to a shapely point
        """
        return shape.contains(_ShapelyPoint((x, y)))

    @bbox.register
    def _(shape: _BaseGeometry):
        """
        If a shape is an array of points, compute the minima/maxima
        or let it pass through if it's 1 dimensional & length 4
        """
        return numpy.asarray(list(shape.bounds))

    @centroid.register
    def _(shape: _BaseGeometry):
        """
        Handle shapely, which requires explicit centroid extraction
        """
        return numpy.asarray(list(shape.centroid.coords)).squeeze()

except ModuleNotFoundError:
    HAS_SHAPELY = False


try:
    import shapely

    from packaging.version import Version

    if Version(shapely.__version__) < Version("2"):
        HAS_SHAPELY2 = False
    else:
        HAS_SHAPELY2 = True

    HULL_TYPES = (*HULL_TYPES, shapely.Geometry)

    @area.register
    def _(shape: shapely.Geometry):
        """
        If we know we're working with a shapely polygon,
        then use shapely.area
        """
        return shapely.area(shape)

    @contains.register
    def _(shape: shapely.Geometry, x: float, y: float):
        """
        If we know we're working with a shapely polygon,
        then use shapely.within casting the points to a shapely object too
        """
        return shapely.within(shapely.points((x, y)), shape)

    @bbox.register
    def _(shape: shapely.Geometry):
        """
        If we know we're working with a shapely polygon,
        then use shapely.bounds
        """
        return shapely.bounds(shape)

    @centroid.register
    def _(shape: shapely.Geometry):
        """
        if we know we're working with a shapely polygon,
        then use shapely.centroid
        """
        return shapely.coordinates.get_coordinates(shapely.centroid(shape)).squeeze()

except ModuleNotFoundError:
    HAS_SHAPELY2 = False

# ------------------------------------------------------------#
# Constructors for trees, prepared inputs, & neighbors        #
# ------------------------------------------------------------#


def build_best_tree(coordinates, metric):
    """
    Build the best query tree that can support the application.
    Chooses from:
    1. sklearn.KDTree if available and metric is simple
    2. sklearn.BallTree if available and metric is complicated

    Parameters
    ----------
    coordinates : numpy.ndarray
        array of coordinates over which to build the tree.
    metric : string or callable
        either a metric supported by sklearn KDTrees or BallTrees, or a callabe function.
        If sklearn is not installed, then this must be euclidean.

    Returns
    -------
        : a distance tree
          a KDTree from either scikit-learn or scipy, or a BallTree if available.

    Notes
    -----
        This will return a scikit-learn KDTree if the metric is supported and
        sklearn can be imported.
        If the metric is not supported by KDTree, a BallTree will be used if
        sklearn can be imported.
        If the metric is a user-defined callable function, a Ball Tree will be used
        if sklearn can be imported.
        If sklearn can't be imported, then a scipy.spatial.KDTree will be used
        if the metric is euclidean.
        Otherwise, an error will be raised.
    """
    coordinates = numpy.asarray(coordinates)
    tree = spatial.KDTree
    try:
        import sklearn
        from sklearn.neighbors import KDTree, BallTree
        from packaging.version import Version

        if Version(sklearn.__version__) == Version("1.3.0"):
            kdtree_valid_metrics = KDTree.valid_metrics()
            balltree_valid_metrics = BallTree.valid_metrics()
        else:
            kdtree_valid_metrics = KDTree.valid_metrics
            balltree_valid_metrics = BallTree.valid_metrics

        if metric in kdtree_valid_metrics:
            tree = lambda coordinates: KDTree(coordinates, metric=metric)
        elif metric in balltree_valid_metrics:
            tree = lambda coordinates: BallTree(coordinates, metric=metric)
        elif callable(metric):
            warnings.warn(
                "Distance metrics defined in pure Python may "
                " have unacceptable performance!",
                stacklevel=2,
            )
            tree = lambda coordinates: BallTree(coordinates, metric=metric)
        else:
            raise KeyError(
                f"Metric {metric} not found in set of available types."
                f"BallTree metrics: {balltree_valid_metrics}, and"
                f"scikit KDTree metrics: {kdtree_valid_metrics}."
            )
    except ModuleNotFoundError as e:
        if metric not in ("l2", "euclidean"):
            raise KeyError(
                f"Metric {metric} requested, but this requires"
                f" scikit-learn to use. Without scikit-learn, only"
                f" euclidean distance metric is supported."
            )
    return tree(coordinates)


def k_neighbors(tree, coordinates, k, **kwargs):
    """
    Query a kdtree for k neighbors, handling the self-neighbor case
    in the case of coincident points.

    Arguments
    ----------
    tree : distance tree
        a distance tree, such as a scipy KDTree or sklearn KDTree or BallTree
        that supports a query argument.
    coordinates : numpy.ndarray of shape n,2
        coordinates to query for their neighbors within the tree.
    k : int
        number of neighbors to query in the tree
    **kwargs : mappable
        arguments that may need to be passed down to the tree.query() function

    Returns
    --------
    a tuple of (distances, indices) that is assured to not include the point itself
    in its query result.
    """
    distances, indices = tree.query(coordinates, k=k + 1, **kwargs)
    n, ks = distances.shape
    assert ks == k + 1
    full_indices = numpy.arange(n)
    other_index_mask = indices != full_indices.reshape(n, 1)
    has_k_indices = other_index_mask.sum(axis=1) == (k + 1)
    other_index_mask[has_k_indices, -1] = False
    distances = distances[other_index_mask].reshape(n, k)
    indices = indices[other_index_mask].reshape(n, k)
    return distances, indices


def prepare_hull(coordinates, hull=None):
    """
    Construct a hull from the coordinates given a hull type
    Will either return:
        - a bounding box array of [xmin, ymin, xmax, ymax]
        - a scipy.spatial.ConvexHull object from the Qhull library
        - a shapely shape using alpha_shape_auto

    Parameters
    ---------
    coordinates : numpy.ndarray of shape (n,2)
        Points to use to construct a hull
    hull : string or a pre-computed hull
        A string denoting what kind of hull to compute (if required) or a hull
        that has already been computed

    Returns
    --------
    hull : A geometry-like object
        This encodes the "space" in which to simulate the normal pattern. All points will
        lie within this hull. Supported values are:
        - a bounding box encoded in a numpy array as numpy.array([xmin, ymin, xmax, ymax])
        - an (N,2) array of points for which the bounding box will be computed & used
        - a shapely polygon/multipolygon
        - a shapely geometry
        - a scipy.spatial.ConvexHull
    """
    if isinstance(hull, numpy.ndarray):
        assert len(hull) == 4, f"bounding box provided is not shaped correctly! {hull}"
        assert hull.ndim == 1, f"bounding box provided is not shaped correctly! {hull}"
        return hull
    if (hull is None) or (hull == "bbox"):
        return bbox(coordinates)
    if HAS_SHAPELY:  # protect the isinstance check if import has failed
        if isinstance(hull, (_ShapelyPolygon, _ShapelyMultiPolygon)):
            return hull
    if HAS_SHAPELY2:
        if isinstance(hull, shapely.Geometry):
            return hull
    if isinstance(hull, str):
        if hull.startswith("convex"):
            return spatial.ConvexHull(coordinates)
        elif hull.startswith("alpha") or hull.startswith("α"):
            return alpha_shape_auto(coordinates)
    elif isinstance(hull, spatial.ConvexHull):
        return hull
    raise ValueError(
        f"Hull type {hull} not in the set of valid options:"
        f" (None, 'bbox', 'convex', 'alpha', 'α', "
        f" shapely.geometry.Polygon, shapely.Geometry)"
    )



import unittest
import numpy as np

from ..pointpattern import PointPattern
from libpysal.common import RTOL


class TestPointPattern(unittest.TestCase):

    def setUp(self):
        self.points = [[66.22, 32.54], [22.52, 22.39], [31.01, 81.21],
                       [9.47, 31.02], [30.78, 60.10], [75.21, 58.93],
                       [79.26,  7.68], [8.23, 39.93], [98.73, 77.17],
                       [89.78, 42.53], [65.19, 92.08], [54.46, 8.48]]
        self.pp = PointPattern(self.points)
        self.assertEqual(len(self.pp), 12)
        self.assertTrue([66.22, 32.54] in self.pp)

    def test_point_pattern_n(self):
        self.assertEqual(self.pp.n, 12)

    def test_point_pattern_mean_nnd(self):
        np.testing.assert_allclose(self.pp.mean_nnd, 21.612139802089246, RTOL)

    def test_point_pattern_lambda_mbb(self):
        np.testing.assert_allclose(self.pp.lambda_mbb,
                                   0.0015710507711240867, RTOL)

    def test_point_pattern_lambda_hull(self):
        np.testing.assert_allclose(self.pp.lambda_hull,
                                   0.0022667153468973137, RTOL)

    def test_point_pattern_hull_area(self):
        np.testing.assert_allclose(self.pp.hull_area, 5294.0039500000003, RTOL)

    def test_point_pattern_mbb_area(self):
        np.testing.assert_allclose(self.pp.mbb_area, 7638.2000000000007, RTOL)

    def test_point_pattern_min_nnd(self):
        np.testing.assert_allclose(self.pp.min_nnd, 8.9958712752017522, RTOL)

    def test_point_pattern_max_nnd(self):
        np.testing.assert_allclose(self.pp.max_nnd, 34.63124167568931, RTOL)

    def test_point_pattern_find_pairs(self):
        self.assertEqual(self.pp.find_pairs(10), {(3, 7)})
        self.assertEqual(self.pp.find_pairs(20), {(3, 7), (1, 3)})

    def test_point_pattern_knn(self):
        knn = self.pp.knn(1)
        nn = np.array([[9], [3], [4], [7], [2], [9], [11], [3], [5], [5], [5],
                      [6]])
        nnd = np.array([[25.59050019], [15.64542745], [21.11125292],
                        [8.99587128], [21.11125292], [21.93729473],
                        [24.81289987], [8.99587128], [29.76387072],
                        [21.93729473], [34.63124168], [24.81289987]])
        np.testing.assert_array_equal(knn[0], nn)
        np.testing.assert_array_almost_equal(knn[1], nnd)

    def test_point_pattern_knn_error(self):
        self.assertRaises(ValueError, self.pp.knn, k=0)

    def test_point_pattern_knn_other(self):
        knn = self.pp.knn_other(self.pp)
        nn = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
        nnd = np.zeros(12)
        np.testing.assert_array_equal(knn[0], nn)
        np.testing.assert_array_equal(knn[1], nnd)

        knn = self.pp.knn_other([0, 0], k=12)
        nn = np.array([1, 3, 7, 11, 4, 0, 6, 2, 5, 9, 10, 8])
        nnd = np.array([31.75629859, 32.43333625, 40.76932425, 55.11625894,
                       67.52346555, 73.78306039, 79.63121247, 86.92919072,
                       95.54731289, 99.34409545, 112.82048794, 125.31090056])
        np.testing.assert_array_equal(knn[0], nn)
        np.testing.assert_array_almost_equal(knn[1], nnd)

    def test_point_pattern_knn_other_error(self):
        knn_other = self.pp.knn_other
        self.assertRaises(ValueError, knn_other, self.pp, k=0)

    def test_point_pattern_explode(self):
        explosion = self.pp.explode('x')
        for ppattern in explosion:
            np.testing.assert_array_equal(ppattern.df.iloc[0],
                                          self.pp.df.loc[ppattern.df.index[0]])

    def test_point_pattern_flip_coordinates(self):
        pp_flip = PointPattern(self.points, coord_names=['The x coordinate',
                                                        'The y coordinate'])
        coord = pp_flip.coord_names
        x_coord, y_coord = pp_flip._x, pp_flip._y
        # Flip the coordinates
        pp_flip.flip_coordinates()

        coord_flipped = pp_flip.coord_names
        x_coord_flipped, y_coord_flipped = pp_flip._x, pp_flip._y
        self.assertEqual(x_coord, y_coord_flipped)
        self.assertEqual(y_coord, x_coord_flipped)
        self.assertEqual(coord, coord_flipped)

        # Flip the coordinates again, they should return to the intial values.
        pp_flip.flip_coordinates()

        coord_again = pp_flip.coord_names
        x_coord_again, y_coord_again = pp_flip._x, pp_flip._y
        self.assertEqual(x_coord, x_coord_again)
        self.assertEqual(y_coord, y_coord_again)
        self.assertEqual(coord, coord_again)
